	/*
    ###########################################################################
    # event processing routines
    ###########################################################################
	*/

#include "GUI.h"

    /*
     ##########################################################################
     # (@)INTERNAL:ProcessEventError(Name, *PerlResult)
     # Pops up a message box in case of error within an event;
     # returns TRUE if errors were, FALSE otherwise, and sets PerlResult
     # according to user's click (CANCEL == -1),
     */
BOOL ProcessEventError(NOTXSPROC char *Name, int* PerlResult) {
    if(strncmp(Name, "main::", 6) == 0) Name += 6;
    if(SvTRUE(ERRSV)) {
        MessageBeep(MB_ICONASTERISK);
        *PerlResult = MessageBox(
            NULL,
            SvPV_nolen(ERRSV),
            Name,
            MB_ICONERROR | MB_OKCANCEL
        );
        if(*PerlResult == IDCANCEL) {
            *PerlResult = -1;
        }
        return TRUE;
    } else {
        return FALSE;
    }
}

    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_Generic(name)
     # Calls an event without arguments;
     # Name must be pre-filled.
     */
int DoEvent_Generic(NOTXSPROC char *Name) {
    int PerlResult;
    int count;
    PerlResult = 1;
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_Generic): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER;
        SAVETMPS;
        PUSHMARK(SP);
        PUTBACK;
        count = perl_call_pv(Name, G_EVAL|G_NOARGS);
        SPAGAIN;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK;
        FREETMPS;
        LEAVE;
    }
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_Generic): returning %d\n", PerlResult);
#endif
    return PerlResult;
}


    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_Long(name,arg)
     # Same as above, but with a long argument.
     */
int DoEvent_Long(NOTXSPROC char *Name, long argh) {
    int PerlResult;
    int count;
    PerlResult = 1;
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_Long): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv(argh)));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
    return PerlResult;
}


    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_TwoLongs(name,arg1,arg2)
     # Same as above, but with two long arguments.
     */
int DoEvent_TwoLongs(NOTXSPROC char *Name, long argone, long argtwo) {
    int PerlResult;
    int count;
    PerlResult = 1;
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_TwoLongs): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv(argone)));
        XPUSHs(sv_2mortal(newSViv(argtwo)));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
    return PerlResult;
}


    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_ButtonClick(name, wparam)
     # calls a toolbar's WM_COMMAND event
     # adds "_ButtonClick" to Name
     */
int DoEvent_ButtonClick(NOTXSPROC char *Name, WPARAM wParam) {
    int PerlResult;
    int count;
    PerlResult = 1;
    strcat(Name, "_ButtonClick");
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_ButtonClick): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv(LOWORD(wParam))));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
    return PerlResult;
}


    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_ListView(name, lparam)
     # calls a listview's item event
     */
int DoEvent_ListView(NOTXSPROC char *Name, LPARAM lParam) {
    int PerlResult;
    int count;
    LPNM_LISTVIEW lv_notify;
    long argh;
    PerlResult = 1;
    lv_notify = (LPNM_LISTVIEW) lParam;
    switch(lv_notify->hdr.code) {
    case LVN_ITEMCHANGED:
        strcat(Name, "_ItemClick");
        argh = (long) lv_notify->iItem;
        break;
    case LVN_COLUMNCLICK:
        strcat(Name, "_ColumnClick");
        argh = (long) lv_notify->iSubItem;
        break;
    }
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_ListView): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv(argh)));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
    return PerlResult;
}

    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_TreeView(name, lparam)
     # calls a treeview's node event
     */
int DoEvent_TreeView(NOTXSPROC char *Name, LPARAM lParam) {
    int PerlResult;
    int count;
    LPNM_TREEVIEW tv_notify;
    PerlResult = 1;
    tv_notify = (LPNM_TREEVIEW) lParam;
    switch(tv_notify->hdr.code) {
    case TVN_SELCHANGED:
        strcat(Name, "_NodeClick");
        break;
    case TVN_ITEMEXPANDED:
        if(tv_notify->action == TVE_COLLAPSE)
            strcat(Name, "_Collapse");
        else
            strcat(Name, "_Expand");
        break;
    case TVN_ITEMEXPANDING:
        if(tv_notify->action == TVE_COLLAPSE)
            strcat(Name, "_Collapsing");
        else
            strcat(Name, "_Expanding");
        break;
    }
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_TreeView): EVENT: %s\n", Name);
#endif
	if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv((long) tv_notify->itemNew.hItem)));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
    return PerlResult;
}

    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_MouseMove(name, lparam, wparam)
     # calls a WM_MOUSEMOVE event
     # adds "_MouseMove" to Name
     */
int DoEvent_MouseMove(NOTXSPROC char *Name, WPARAM wParam, LPARAM lParam) {
    int PerlResult;
    int count;
    PerlResult = 1;
    strcat(Name, "_MouseMove");
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_MouseMove): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv(wParam)));
        XPUSHs(sv_2mortal(newSViv(LOWORD(lParam))));
        XPUSHs(sv_2mortal(newSViv(HIWORD(lParam))));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
    return PerlResult;
}

    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_MouseButton(name, wparam, lparam)
     # calls a WM_(L/R)BUTTON(UP/DOWN) event
     # Name must be pre-filled
     */
int DoEvent_MouseButton(NOTXSPROC char *Name, WPARAM wParam, LPARAM lParam) {
    int PerlResult;
    int count;
    PerlResult = 1;
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_MouseButton): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv(wParam)));
        XPUSHs(sv_2mortal(newSViv(LOWORD(lParam))));
        XPUSHs(sv_2mortal(newSViv(HIWORD(lParam))));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) PerlResult = POPi;
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
    return PerlResult;
}


    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_NeetText(name, id)
     # calls a TTN_NEEDTEXT event ("callback"?)
     # adds "_NeedText" to Name
     */
char * DoEvent_NeedText(NOTXSPROC char *Name, UINT id) {
    int PerlResult;
    static char *textneeded;
    SV* svt;
    int count;
    strcat(Name, "_NeedText");
    if(textneeded != NULL) {
        safefree(textneeded);
        textneeded = NULL;
    }
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_NeedText): EVENT: %s\n", Name);
#endif
    if(perl_get_cv(Name, FALSE) != NULL) {
        dSP;
        dTARG;
        ENTER ;
        SAVETMPS;
        PUSHMARK(SP) ;
        XPUSHs(sv_2mortal(newSViv(id)));
        PUTBACK ;
        count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        SPAGAIN ;
        if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            if(count > 0) {
                svt = POPs;
                textneeded = (char *) safemalloc(sv_len(svt));
                strcpy(textneeded, SvPV_nolen(svt));
            } else {
#ifdef PERLWIN32GUI_STRONGDEBUG
                printf("!XS(DoEvent_NeedText): sub returned nothing\n");
#endif
            }
        }
        PUTBACK ;
        FREETMPS ;
        LEAVE ;
    }
#ifdef PERLWIN32GUI_DEBUG
    printf("!XS(DoEvent_NeedText): returning '%s'\n", textneeded);
#endif
    return textneeded;
}


    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_NEM(perlud, event_id, name, ...)
     */
int DoEvent_NEM(
	NOTXSPROC
	LPPERLWIN32GUI_USERDATA perlud,
	int iEventId,
	char *Name,
	...
) {
    int PerlResult;
    int count;
    int argtype;
    SV** event;
    va_list args;
    va_start( args, Name );
    PerlResult = -2;
	if( ValidUserData(perlud) ) {
#ifdef PERLWIN32GUI_STRONGDEBUG
	    printf("!XS(DoEvent_NEM) window='%s', event='%s', event_id=0x%x, event_mask=0x%x\n", perlud->szWindowName, Name, iEventId, perlud->dwEventMask);
#endif
		if(perlud->dwEventMask & iEventId) {
			event = hv_fetch( (perlud->hvEvents), Name, strlen(Name), 0);
			if(event != NULL) {
#ifdef PERLWIN32GUI_STRONGDEBUG
			    printf("!XS(DoEvent_NEM) CV found (%s)\n", SvPV_nolen(*event));
#endif
				dSP;
				dTARG;
				ENTER;
				SAVETMPS;
				PUSHMARK(SP);
				XPUSHs(perlud->svSelf);
				argtype = va_arg( args, int );
				while(argtype != -1) {
					switch(argtype) {
					case PERLWIN32GUI_ARGTYPE_INT:
						XPUSHs(sv_2mortal(newSViv(va_arg( args, int ))));
						break;
					case PERLWIN32GUI_ARGTYPE_LONG:
						XPUSHs(sv_2mortal(newSViv(va_arg( args, long ))));
						break;
					case PERLWIN32GUI_ARGTYPE_WORD:
						XPUSHs(sv_2mortal(newSViv(va_arg( args, WORD ))));
						break;
					case PERLWIN32GUI_ARGTYPE_STRING:
						XPUSHs(sv_2mortal(newSVpv(va_arg( args, char * ), 0)));
						break;
					default:
						warn("Win32::GUI: WARNING! unknown argument type (%d) to event '%s'", argtype, Name);
						break;
					}
					argtype = va_arg( args, int );
				}
				va_end( args );
				PUTBACK;
				count = call_sv(*event, G_EVAL|G_ARRAY);
				SPAGAIN;
				if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
					if(count > 0) PerlResult = POPi;
				}
				PUTBACK;
				FREETMPS;
				LEAVE;
			}
		}
    }
    va_end(args);
    return PerlResult;
}

    /*
     ##########################################################################
     # (@)INTERNAL:DoEvent_NEM_Menu(nID)
     */
int DoEvent_NEM_Menu(
	NOTXSPROC
	HWND hwnd,
	int nID,
	...
) {
    int PerlResult;
    int count;
    int argtype;
	SV* event;
    PerlResult = -2;

	event = &PL_sv_undef;
	MENUITEMINFO mii;
	HMENU hmenu;
	LPPERLWIN32GUI_MENUITEMDATA perlmid;
	ZeroMemory(&mii, sizeof(MENUITEMINFO));
	mii.cbSize = sizeof(MENUITEMINFO);
	mii.fMask = MIIM_DATA;
	/* HEURISTIC: assume the message was from the window's own menu */
	hmenu = GetMenu(hwnd);
	/* HEURISTIC: no, it wasn't, search in Perl's global hash  */
	if(hmenu == NULL) hmenu = GetMenuFromID( NOTXSCALL nID );
	/* HEURISTIC: if we can get to the item, it's ok, otherwise search in Perl's global hash  */
    if(GetMenuItemInfo( hmenu, nID, 0, &mii ) == 0) {
		hmenu = GetMenuFromID( NOTXSCALL nID );
	}
    if(GetMenuItemInfo( hmenu, nID, 0, &mii )) {
		perlmid = (LPPERLWIN32GUI_MENUITEMDATA) mii.dwItemData;
		if(perlmid != NULL && perlmid->dwSize == sizeof(PERLWIN32GUI_MENUITEMDATA)) {
			event = perlmid->svCode;
		}
	}

	if( event != &PL_sv_undef ) {
#ifdef PERLWIN32GUI_STRONGDEBUG
	    printf("!XS(DoEvent_NEM_Menu) CV found (%s)\n", SvPV_nolen(event));
#endif
		dSP;
		dTARG;
		ENTER;
		SAVETMPS;
		PUSHMARK(SP);
/*		XPUSHs(perlmid->svSelf); */
		PUTBACK;
		count = call_sv(event, G_EVAL|G_ARRAY);
		SPAGAIN;
		if(!ProcessEventError(NOTXSCALL "", &PerlResult)) {
			if(count > 0) PerlResult = POPi;
		}
		PUTBACK;
		FREETMPS;
		LEAVE;
	}
	return PerlResult;
}
