	/*
    ###########################################################################
    # message loops
    ###########################################################################
	*/

#include "GUI.h"

    /*
    ###########################################################################
    # (@)INTERNAL:CommonMsgLoop(hwnd, uMsg, wParam, lParam)
    # this is the message loop (WndProc) that process common messages
    # (eventmodel independent)
    */
LRESULT CommonMsgLoop(NOTXSPROC HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
#ifdef PERLWIN32GUI_STRONGDEBUG
	    printf("!XS(CommonMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
	switch(uMsg) {

    case WM_CTLCOLOREDIT:
    case WM_CTLCOLORSTATIC:
    case WM_CTLCOLORBTN:
    case WM_CTLCOLORLISTBOX:
        {
			LPPERLWIN32GUI_USERDATA perlud;

			/*
            if(uMsg == WM_CTLCOLORSTATIC
            && GetWindowLong((HWND) lParam, GWL_STYLE) & SS_SIMPLE) {
				FREETMPS;
				LEAVE;
                return FALSE;
            }
            */

			perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong((HWND) lParam, GWL_USERDATA);
			if( ! ValidUserData(perlud) ) {
				HBRUSH defBrush;
				switch(uMsg) {
				case WM_CTLCOLOREDIT:
				case WM_CTLCOLORLISTBOX:
					defBrush = GetSysColorBrush(COLOR_WINDOW);
					break;
				default:
					defBrush = GetSysColorBrush(COLOR_BTNFACE);
					break;
				}
				return ((LRESULT) defBrush);
			}

			if(uMsg == WM_CTLCOLORSTATIC) SetBkMode((HDC) wParam, TRANSPARENT);
			if(perlud->clrForeground != CLR_INVALID) {
				SetTextColor((HDC) wParam, perlud->clrForeground);
			}
			if(perlud->clrBackground != CLR_INVALID) {
				SetBkColor((HDC) wParam, perlud->clrBackground);
				return ((LRESULT) perlud->hBackgroundBrush);
			} else {
				HBRUSH defBrush;
				switch(uMsg) {
				case WM_CTLCOLOREDIT:
				case WM_CTLCOLORLISTBOX:
					defBrush = GetSysColorBrush(COLOR_WINDOW);
					break;
				default:
					defBrush = GetSysColorBrush(COLOR_BTNFACE);
					break;
				}
				return ((LRESULT) defBrush);
			}
        }
        break;

    case WM_GETMINMAXINFO:
        {
            LPMINMAXINFO minmax;
            LPPERLWIN32GUI_USERDATA perlud;
            perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA);
            if( ValidUserData(perlud) ) {
	            minmax = (LPMINMAXINFO) lParam;
				if(perlud->iMinWidth != -1) minmax->ptMinTrackSize.x = (LONG) perlud->iMinWidth;
				if(perlud->iMaxWidth != -1) minmax->ptMaxTrackSize.x = (LONG) perlud->iMaxWidth;
				if(perlud->iMinHeight != -1) minmax->ptMinTrackSize.y = (LONG) perlud->iMinHeight;
				if(perlud->iMaxHeight != -1) minmax->ptMaxTrackSize.y = (LONG) perlud->iMaxHeight;
				return 0;
			}
		}
        break;

    case WM_SETCURSOR:
        {
            LPPERLWIN32GUI_USERDATA perlud;
            perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong((HWND) wParam, GWL_USERDATA);
            if( ValidUserData(perlud) ) {
				if( perlud->hCursor != NULL) {
					SetCursor( perlud->hCursor );
					return TRUE;
				} else {
					return FALSE;
				}
			}
			return FALSE;
		}
        break;

	}
#ifdef PERLWIN32GUI_STRONGDEBUG
	printf("!XS(CommonMsgLoop) returning DefWindowProc\n");
#endif
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}


    /*
    ###########################################################################
    # (@)INTERNAL:NEM_WindowMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for the New Event Model (NEM)
    */
LRESULT CALLBACK NEM_WindowMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    HWND hwndParent;
	LPPERLWIN32GUI_USERDATA perlud;
	LPPERLWIN32GUI_USERDATA childud;
	WNDPROC wndprocOriginal;
	int PerlResult;

#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(NEM_WindowMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
    if(uMsg == WM_CREATE || uMsg == WM_NCCREATE) {
        perlud = (LPPERLWIN32GUI_USERDATA) ((CREATESTRUCT *) lParam)->lpCreateParams;
        if(perlud!= NULL) {
            SetWindowLong(hwnd, GWL_USERDATA, (long) perlud);
        }
#ifdef PERLWIN32GUI_STRONGDEBUG
		printf("!XS(NEM_WindowMsgLoop) returning DefWindowProc (WM_CREATE)\n");
#endif
        return DefWindowProc(hwnd, uMsg, wParam, lParam);
    }
    if(uMsg == WM_DESTROY) {
		safefree( (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA));
#ifdef PERLWIN32GUI_STRONGDEBUG
		printf("!XS(NEM_WindowMsgLoop) returning DefWindowProc (WM_DESTROY)\n");
#endif
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}

	PerlResult = 1;
	perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA);
	if( ! ValidUserData(perlud) ) {
#ifdef PERLWIN32GUI_STRONGDEBUG
	    printf("!XS(NEM_WindowMsgLoop) returning DefWindowProc (perlud not valid)\n");
#endif
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
#ifdef PERL_OBJECT
    CPerl *pPerl;
	pPerl = perlud->pPerl;
#endif
	switch(uMsg) {


    case WM_ACTIVATE:
    	if(perlud->dwPlStyle & PERLWIN32GUI_NEM) {
            if(LOWORD(wParam) == WA_INACTIVE) {
				PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_DEACTIVATE, "Deactivate", -1 );
            } else {
				PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_ACTIVATE, "Activate", -1 );
            }
        }
        break;

    case WM_SYSCOMMAND:
    	if(perlud->dwPlStyle & PERLWIN32GUI_NEM) {
            switch(wParam & 0xFFF0) {
            case SC_CLOSE:
            	PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_TERMINATE, "Terminate", -1 );
                break;
            case SC_MINIMIZE:
            	PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_MINIMIZE, "Minimize", -1 );
                break;
            case SC_MAXIMIZE:
            	PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_MAXIMIZE, "Maximize", -1 );
                break;
            }
        }
        break;

    case WM_SIZE:
    	if(perlud->dwPlStyle & PERLWIN32GUI_NEM) {
           	PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_RESIZE, "Resize", -1 );
        }
        break;

	case WM_COMMAND:
		if(HIWORD(wParam) == 0 && lParam == NULL) {
			// menu command processing
#ifdef PERLWIN32GUI_STRONGDEBUG
			printf("!XS(NEM_WindowMsgLoop) got WM_COMMAND for a menu...\n");
#endif
			PerlResult = DoEvent_NEM_Menu(NOTXSCALL hwnd, LOWORD(wParam));
		} else if(HIWORD(wParam) == 1 && lParam == NULL) {
			// accelerator processing
			/*
#ifdef PERLWIN32GUI_STRONGDEBUG
			printf("!XS(WindowMsgLoop) got WM_COMMAND for an accelerator...\n");
#endif
			if(GetAcceleratorName(NOTXSCALL LOWORD(wParam), Name)) {
				PerlResult = DoEvent_NEM_Acc(NOTXSCALL accud);
			}
			*/
		} else {

			childud = (LPPERLWIN32GUI_USERDATA) GetWindowLong((HWND) lParam, GWL_USERDATA);
			if( ValidUserData(childud) && (childud->dwPlStyle & PERLWIN32GUI_NEM) ) {
				switch(childud->iClass) {
				case PERLWIN32GUI_CLS_BUTTON:
				case PERLWIN32GUI_CLS_CHECKBOX:
				case PERLWIN32GUI_CLS_RADIOBUTTON:
					switch(HIWORD(wParam)) {
					case BN_SETFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_GOTFOCUS, "GotFocus", -1 );
						break;
					case BN_KILLFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_LOSTFOCUS, "LostFocus", -1 );
						break;
					case BN_DBLCLK:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_DBLCLICK, "DblClick", -1 );
						break;
					case BN_CLICKED:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_CLICK, "Click", -1 );
						break;
					}
					break;
				case PERLWIN32GUI_CLS_LISTBOX:
					switch(HIWORD(wParam)) {
					case LBN_SETFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_GOTFOCUS, "GotFocus", -1 );
						break;
					case LBN_KILLFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_LOSTFOCUS, "LostFocus", -1 );
						break;
					case LBN_DBLCLK:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_DBLCLICK, "DblClick", -1 );
						break;
					case LBN_SELCHANGE:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_CLICK, "Click", -1 );
						break;
					}
					break;
				case PERLWIN32GUI_CLS_EDIT:
				case PERLWIN32GUI_CLS_RICHEDIT:
					switch(HIWORD(wParam)) {
					case EN_SETFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_GOTFOCUS, "GotFocus", -1 );
						break;
					case EN_KILLFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_LOSTFOCUS, "LostFocus", -1 );
						break;
					case EN_CHANGE:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_CHANGE, "Change", -1 );
						break;
					}
					break;
				case PERLWIN32GUI_CLS_STATIC:
					switch(HIWORD(wParam)) {
					case STN_CLICKED:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_CLICK, "Click", -1 );
						break;
					case STN_DBLCLK:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_DBLCLICK, "DblClick", -1 );
						break;
					}
					break;
				case PERLWIN32GUI_CLS_COMBOBOX:
					switch(HIWORD(wParam)) {
					case CBN_SETFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_GOTFOCUS, "GotFocus", -1 );
						break;
					case CBN_KILLFOCUS:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_LOSTFOCUS, "LostFocus", -1 );
						break;
					case CBN_DBLCLK:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_DBLCLICK, "DblClick", -1 );
						break;
					case CBN_SELCHANGE:
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_CLICK, "Change", -1 );
						break;
					}
					break;
				case PERLWIN32GUI_CLS_TOOLBAR:
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_BUTTONCLICK, "ButtonClick",
						PERLWIN32GUI_ARGTYPE_WORD, LOWORD(wParam), -1
					);
					break;
				}
			}
		}
		break;
	case WM_NOTIFY:
		LPNMHDR notify;
		notify = (LPNMHDR) lParam;
		childud = (LPPERLWIN32GUI_USERDATA) GetWindowLong(notify->hwndFrom, GWL_USERDATA);
		if( ValidUserData(childud) && (childud->dwPlStyle & PERLWIN32GUI_NEM) ) {
			switch(childud->iClass) {
			case PERLWIN32GUI_CLS_LISTVIEW:
				LPNM_LISTVIEW lv_notify;
				lv_notify = (LPNM_LISTVIEW) lParam;
#ifdef PERLWIN32GUI_STRONDEBUG
				printf("!XS(NEM_WindowMsgLoop) WM_NOTIFY on ListView, code=0x%d\n", notify->code);
#endif
				switch(notify->code) {
				case LVN_ITEMCHANGED:
					if(lv_notify->uChanged & LVIF_STATE
					&& lv_notify->uNewState & LVIS_SELECTED) {
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_ITEMCLICK, "ItemClick",
							PERLWIN32GUI_ARGTYPE_INT, lv_notify->iItem, -1
						);
					}
					if(lv_notify->uChanged & LVIF_STATE
					&& (lv_notify->uOldState & LVIS_STATEIMAGEMASK) != (lv_notify->uNewState & LVIS_STATEIMAGEMASK)) {
						PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_ITEMCHECK, "ItemCheck",
							PERLWIN32GUI_ARGTYPE_INT, lv_notify->iItem, -1
						);
					}
					break;
				case LVN_COLUMNCLICK:
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_COLUMNCLICK, "ColumnClick",
						PERLWIN32GUI_ARGTYPE_INT, lv_notify->iSubItem, -1
					);
					break;
				case LVN_KEYDOWN:
					LV_KEYDOWN FAR * lv_keydown;
					lv_keydown = (LV_KEYDOWN FAR *) lParam;
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_KEYDOWN, "KeyDown",
						PERLWIN32GUI_ARGTYPE_WORD, lv_keydown->wVKey, -1
					);
					break;
				}
                break;
			case PERLWIN32GUI_CLS_TREEVIEW:
				LPNM_TREEVIEW tv_notify;
                tv_notify = (LPNM_TREEVIEW) lParam;
                switch(notify->code) {
                case TVN_ITEMEXPANDED:
                    if(tv_notify->action == TVE_COLLAPSE) {
                    	PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_COLLAPSE, "Collapse",
                    		PERLWIN32GUI_ARGTYPE_LONG, tv_notify->itemNew.hItem, -1
                    	);
                    } else {
                    	PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_EXPAND, "Expand",
                    		PERLWIN32GUI_ARGTYPE_LONG, tv_notify->itemNew.hItem, -1
						);
					}
                    break;
                case TVN_ITEMEXPANDING:
                    if(tv_notify->action == TVE_COLLAPSE) {
                    	PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_COLLAPSING, "Collapsing",
                    		PERLWIN32GUI_ARGTYPE_LONG, tv_notify->itemNew.hItem, -1
						);
                    } else {
                    	PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_EXPANDING, "Expanding",
                    		PERLWIN32GUI_ARGTYPE_LONG, tv_notify->itemNew.hItem, -1
						);
					}
                    FREETMPS;
                    LEAVE;
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                    break;
                case TVN_SELCHANGED:
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_NODECLICK, "NodeClick",
						PERLWIN32GUI_ARGTYPE_LONG, tv_notify->itemNew.hItem, -1
					);
                    break;
                case TVN_KEYDOWN:
                	TV_KEYDOWN FAR * tv_keydown;
                    tv_keydown = (TV_KEYDOWN FAR *) lParam;
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_KEYDOWN, "KeyDown",
						PERLWIN32GUI_ARGTYPE_WORD, tv_keydown->wVKey, -1
					);
                    break;
                }
                break;
            case PERLWIN32GUI_CLS_TAB:
                switch(notify->code) {
                case TCN_SELCHANGING:
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_CHANGING, "Changing", -1);
                    FREETMPS;
                    LEAVE;
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                    break;
                case TCN_SELCHANGE:
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_CHANGE, "Change", -1);
                    break;
                }
                break;

            case PERLWIN32GUI_CLS_TOOLTIP:
                switch(notify->code) {
                case TTN_NEEDTEXT:
                    LPTOOLTIPTEXT lptt;
                    lptt = (LPTOOLTIPTEXT) lParam;
                    lptt->lpszText = (LPTSTR) DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_NEEDTEXT, "NeedText",
                    	PERLWIN32GUI_ARGTYPE_INT, lptt->hdr.idFrom, -1
                    );
#ifdef PERLWIN32GUI_STRONGDEBUG
	                printf("!XS(NEM_WindowMsgLoop): TTN_NEEDTEXT got '%s'\n", lptt->lpszText);
#endif
                    PerlResult = 1;
					break;
                }
                break;

            case PERLWIN32GUI_CLS_REBAR:
                switch(notify->code) {
                case RBN_HEIGHTCHANGE:
					PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_HEIGHTCHANGE, "HeightChange", -1);
					break;
                }
                break;

            case PERLWIN32GUI_CLS_HEADER:
				LPNMHEADER nmh;
				nmh = (LPNMHEADER) lParam;
                switch(notify->code) {
                case HDN_BEGINTRACK:
                    PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_BEGINTRACK, "BeginTrack",
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->iItem,
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->pitem->cxy,
                    	-1
                    );
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                    break;
                case HDN_ENDTRACK:
                    PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_ENDTRACK, "EndTrack",
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->iItem,
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->pitem->cxy,
                    	-1
                    );
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
					break;
                case HDN_TRACK:
                    PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_TRACK, "Track",
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->iItem,
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->pitem->cxy,
                    	-1
                    );
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                    break;
                case HDN_DIVIDERDBLCLICK:
                    PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_DIVIDERDBLCLICK, "DividerDblClick",
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->iItem,
                    	-1
                    );
                    break;
                case HDN_ITEMCLICK:
                    PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_ITEMCLICK, "ItemClick",
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->iItem,
                    	-1
                    );
                    break;
                case HDN_ITEMDBLCLICK:
                    PerlResult = DoEvent_NEM(NOTXSCALL childud, PERLWIN32GUI_NEM_ITEMDBLCLICK, "ItemDblClick",
                    	PERLWIN32GUI_ARGTYPE_INT, nmh->iItem,
                    	-1
                    );
                    break;
                }
                break;
			}
		}
		break;
	}
    if(PerlResult == -1) {
        PostMessage(hwnd, WM_EXITLOOP, -1, 0);
        return 0;
    } else if(PerlResult == 0) {
        return 0;
    } else if(PerlResult == -2) { /* DoEvent_NEM did not find a suitable event */
		if(perlud->iEventModel == PERLWIN32GUI_EM_BOTH) {
			if(perlud->wndprocPreNEM != NULL) {
				wndprocOriginal = perlud->wndprocPreNEM;
				return CallWindowProc((FARPROC) wndprocOriginal, hwnd, uMsg, wParam, lParam);
			} else {
	        	return CommonMsgLoop(NOTXSCALL hwnd, uMsg, wParam, lParam);
			}
		} else {
			return CommonMsgLoop(NOTXSCALL hwnd, uMsg, wParam, lParam);
		}
	} else {
        return CommonMsgLoop(NOTXSCALL hwnd, uMsg, wParam, lParam);
	}
}


    /*
    ###########################################################################
    # (@)INTERNAL:NEM_ChildMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for the New Event Model (NEM)
    */
LRESULT CALLBACK NEM_ChildMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    HWND hwndParent;
	LPPERLWIN32GUI_USERDATA perlud;
	WNDPROC wndprocOriginal;
	int PerlResult;
#ifdef PERLWIN32GUI_STRONGDEBUG
	printf("!XS(NEM_ChildMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
    if(uMsg == WM_CREATE || uMsg == WM_NCCREATE) {
        perlud = (LPPERLWIN32GUI_USERDATA) ((CREATESTRUCT *) lParam)->lpCreateParams;
        if(perlud!= NULL) {
            SetWindowLong(hwnd, GWL_USERDATA, (long) perlud);
        }
#ifdef PERLWIN32GUI_STRONGDEBUG
		printf("!XS(NEM_ChildMsgLoop) returning DefWindowProc (WM_CREATE)\n");
#endif
        return DefWindowProc(hwnd, uMsg, wParam, lParam);
    }
    if(uMsg == WM_DESTROY) {
		safefree( (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA));
#ifdef PERLWIN32GUI_STRONGDEBUG
		printf("!XS(NEM_ChildMsgLoop) returning DefWindowProc (WM_DESTROY)\n");
#endif
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}

	PerlResult = 1;
	perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA);
	if( ! ValidUserData(perlud) ) {
#ifdef PERLWIN32GUI_STRONGDEBUG
    	printf("!XS(NEM_ChildMsgLoop) returning DefWindowProc (perlud not valid)\n");
#endif
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
#ifdef PERL_OBJECT
	CPerl *pPerl;
	pPerl = perlud->pPerl;
#endif
	switch(uMsg) {
	case WM_MOUSEMOVE:
		PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_MOUSEMOVE, "MouseMove",
			PERLWIN32GUI_ARGTYPE_LONG, LOWORD(lParam),
			PERLWIN32GUI_ARGTYPE_LONG, HIWORD(lParam),
			PERLWIN32GUI_ARGTYPE_LONG, wParam,
		-1);
        break;
	case WM_LBUTTONDOWN:
		PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_MOUSEDOWN, "MouseDown",
			PERLWIN32GUI_ARGTYPE_LONG, LOWORD(lParam),
			PERLWIN32GUI_ARGTYPE_LONG, HIWORD(lParam),
			PERLWIN32GUI_ARGTYPE_LONG, wParam,
		-1);
        break;
	case WM_LBUTTONUP:
		PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_MOUSEUP, "MouseUp",
			PERLWIN32GUI_ARGTYPE_LONG, LOWORD(lParam),
			PERLWIN32GUI_ARGTYPE_LONG, HIWORD(lParam),
			PERLWIN32GUI_ARGTYPE_LONG, wParam,
		-1);
        break;
	case WM_KEYDOWN:
		PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_KEYDOWN, "KeyDown", -1);
        break;
	case WM_KEYUP:
		PerlResult = DoEvent_NEM(NOTXSCALL perlud, PERLWIN32GUI_NEM_KEYUP, "KeyUp", -1);
        break;
    }

    if(PerlResult == 0) {
        return 0;
    } else {
		if(perlud->wndprocPreNEM != NULL) {
			wndprocOriginal = perlud->wndprocPreNEM;
#ifdef PERLWIN32GUI_STRONGDEBUG
			printf("!XS(NEM_ChildMsgLoop) calling wndprocOriginal\n");
#endif
			return CallWindowProc((FARPROC) wndprocOriginal, hwnd, uMsg, wParam, lParam);
		} else {
#ifdef PERLWIN32GUI_STRONGDEBUG
			printf("!XS(NEM_ChildMsgLoop) returning DefWindowProc (wndprocOriginal not valid)\n");
#endif
        	return DefWindowProc(hwnd, uMsg, wParam, lParam);
		}
    }
}


    /*
    ###########################################################################
    # (@)INTERNAL:ContainerMsgLoop(hwnd, uMsg, wParam, lParam)
    */
LRESULT CALLBACK ContainerMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    HWND hwndParent;
	LPPERLWIN32GUI_USERDATA perlud;
	WNDPROC wndprocOriginal;

#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(ContainerMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif

    if(uMsg == WM_COMMAND || uMsg == WM_NOTIFY) {
#ifdef PERLWIN32GUI_STRONGDEBUG
		printf("!XS(ContainerMsgLoop) got WM_COMMAND/WM_NOTIFY, forwarding to parent...\n");
#endif
        hwndParent = (HWND) GetWindowLong(hwnd, GWL_HWNDPARENT);
        SendMessage(hwndParent, uMsg, wParam, lParam);
        return 0;
    } else {
		perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA);
		if( ValidUserData(perlud) && (perlud->wndprocPreContainer != NULL) ) {
			wndprocOriginal = perlud->wndprocPreContainer;
#ifdef PERLWIN32GUI_STRONGDEBUG
			printf("!XS(ContainerMsgLoop) perlud was valid, forwarding to wndprocOriginal(0x%x)...\n", wndprocOriginal);
#endif
			return CallWindowProc((FARPROC) wndprocOriginal, hwnd, uMsg, wParam, lParam);
		} else {
#ifdef PERLWIN32GUI_STRONGDEBUG
			printf("!XS(ContainerMsgLoop) perlud was not valid, forwarding to DefWindowProc...\n");
#endif
			return DefWindowProc(hwnd, uMsg, wParam, lParam);
		}
    }
}


    /*
    ###########################################################################
    # (@)INTERNAL:SplitterMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for Win32::GUI::Splitter objects
    */
LRESULT CALLBACK SplitterMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    int PerlResult;
    char Name[MAX_EVENT_NAME];
    HV* self;
    SV** fetching;
    SV* storing;
    BOOL tracking, horizontal;
    int min;
    int max;
    POINT pt;
    HWND phwnd;
    HDC hdc;
    RECT rc;
	LPPERLWIN32GUI_USERDATA perlud;
	BOOL perludValid;

	perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA);
	if( ValidUserData(perlud) ) {
		perludValid = TRUE;
	} else {
		perludValid = FALSE;
	}
#ifdef PERL_OBJECT
	if(perludValid) {
	    CPerl *pPerl;
		pPerl = perlud->pPerl;
	} else {
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
#endif

#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(SplitterMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif

    if(uMsg == WM_DESTROY) {
		if(perludValid) safefree( perlud );
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}

    PerlResult = 1;
    strcpy(Name, "main::");
    if(GetObjectName(NOTXSCALL hwnd, Name)) {
        switch(uMsg) {
        case WM_MOUSEMOVE:
			if(perludValid) {
				tracking = perlud->dwPlStyle & PERLWIN32GUI_TRACKING;
				if(tracking) {
					horizontal = perlud->dwPlStyle & PERLWIN32GUI_HORIZONTAL;
					if(horizontal) {
						phwnd = GetParent(hwnd);
						GetCursorPos(&pt);
						ScreenToClient(phwnd, &pt);
						pt.y = AdjustSplitterCoord(NOTXSCALL perlud, pt.y, phwnd);
						DrawSplitter(NOTXSCALL hwnd);
						GetWindowRect(hwnd, &rc);
						ScreenToClient(phwnd, (POINT*)&rc);
						SetWindowPos(hwnd, NULL, rc.left, pt.y, 0, 0, SWP_NOZORDER | SWP_NOSIZE);
						DrawSplitter(NOTXSCALL hwnd);
					} else {
						phwnd = GetParent(hwnd);
						GetCursorPos(&pt);
						ScreenToClient(phwnd, &pt);
						pt.x = AdjustSplitterCoord(NOTXSCALL perlud, pt.x, phwnd);
						DrawSplitter(NOTXSCALL hwnd);
						GetWindowRect(hwnd, &rc);
						ScreenToClient(phwnd, (POINT*)&rc);
						SetWindowPos(hwnd, NULL, pt.x, rc.top, 0, 0, SWP_NOZORDER | SWP_NOSIZE);
						DrawSplitter(NOTXSCALL hwnd);
					}
				}
			}
            break;
        case WM_LBUTTONDOWN:
			if(perludValid) {
				SwitchBit(perlud->dwPlStyle, PERLWIN32GUI_TRACKING, 1);
				SetWindowLong(hwnd, GWL_USERDATA, (LONG) perlud);
				horizontal = perlud->dwPlStyle & PERLWIN32GUI_HORIZONTAL;
				if(horizontal) {
					phwnd = GetParent(hwnd);
					GetCursorPos(&pt);
					ScreenToClient(phwnd, &pt);
					pt.y = AdjustSplitterCoord(NOTXSCALL perlud, pt.y, phwnd);
					DrawSplitter(NOTXSCALL hwnd);
					SetCapture(hwnd);
				} else {
					phwnd = GetParent(hwnd);
					GetCursorPos(&pt);
					ScreenToClient(phwnd, &pt);
					pt.x = AdjustSplitterCoord(NOTXSCALL perlud, pt.x, phwnd);
					DrawSplitter(NOTXSCALL hwnd);
					SetCapture(hwnd);
				}
			}
			break;
		case WM_LBUTTONUP:
        	if(perludValid) {
				tracking = perlud->dwPlStyle & PERLWIN32GUI_TRACKING;
				if(tracking) {
					horizontal = perlud->dwPlStyle & PERLWIN32GUI_HORIZONTAL;
					/*
					 * (@)EVENT:Release(COORD)
					 * Sent when the Splitter is released after being
					 * dragged to a new location (identified by the
					 * COORD parameter).
					 * (@)APPLIES_TO:Splitter
					 */
					if(horizontal) {
						phwnd = GetParent(hwnd);
						GetCursorPos(&pt);
						ScreenToClient(phwnd, &pt);
						pt.y = AdjustSplitterCoord(NOTXSCALL perlud, pt.y, phwnd);
						DrawSplitter(NOTXSCALL hwnd);
						strcat(Name, "_Release");
						PerlResult = DoEvent_Long(NOTXSCALL Name, (long) pt.y);
					} else {
						phwnd = GetParent(hwnd);
						GetCursorPos(&pt);
						ScreenToClient(phwnd, &pt);
						pt.x = AdjustSplitterCoord(NOTXSCALL perlud, pt.x, phwnd);
						DrawSplitter(NOTXSCALL hwnd);
						strcat(Name, "_Release");
						PerlResult = DoEvent_Long(NOTXSCALL Name, (long) pt.x);
					}
				}
				SwitchBit(perlud->dwPlStyle, PERLWIN32GUI_TRACKING, 0);
				SetWindowLong(hwnd, GWL_USERDATA, (LONG) perlud);
			}
			ReleaseCapture();
            break;
        }
    }
    if(PerlResult == 0) {
        return 0;
    } else {
        return DefWindowProc(hwnd, uMsg, wParam, lParam);
    }
}

    /*
    ###########################################################################
    # (@)INTERNAL:ButtonMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for subclassed Win32::GUI::Button objects
    */
LRESULT CALLBACK ButtonMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
#ifdef PERL_OBJECT
    CPerl *pPerl;
    pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
	if(pPerl == NULL) return DefButtonProc(hwnd, uMsg, wParam, lParam);
#endif
#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(ButtonMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
    int PerlResult;
    char Name[MAX_EVENT_NAME];
    PerlResult = 1;
    strcpy(Name, "main::");
    if(GetObjectName(NOTXSCALL hwnd, Name)) {
        switch(uMsg) {
        case WM_MOUSEMOVE:
            PerlResult = DoEvent_MouseMove(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_MOUSEHOVER:
        	strcat((char *) Name, "_MouseOver");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;
        case WM_MOUSELEAVE:
        	strcat((char *) Name, "_MouseOut");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;

        // to implement:
        // MouseUp
        // MouseDown
        // KeyPress
        }
    }
    if(PerlResult == 0) {
        return 0;
    } else {
        return DefButtonProc(hwnd, uMsg, wParam, lParam);
    }
}

    /*
    ###########################################################################
    # (@)INTERNAL:ListboxMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for subclassed Win32::GUI::Listbox objects
    */
LRESULT CALLBACK ListboxMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
#ifdef PERL_OBJECT
    CPerl *pPerl;
	pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
	if(pPerl == NULL) return DefListboxProc(hwnd, uMsg, wParam, lParam);
#endif
#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(ListboxMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
    int PerlResult;
    char Name[MAX_EVENT_NAME];
    PerlResult = 1;
    strcpy((char *) Name, "main::");
    if(GetObjectName(NOTXSCALL hwnd, Name)) {
        switch(uMsg) {
        case WM_MOUSEMOVE:
            PerlResult = DoEvent_MouseMove(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_MOUSEHOVER:
        	strcat((char *) Name, "_MouseOver");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;
        case WM_MOUSELEAVE:
        	strcat((char *) Name, "_MouseOut");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;

        // to implement:
        // MouseUp
        // MouseDown
        // KeyPress
        }
    }
    if(PerlResult == 0) {
        return 0;
    } else {
        return DefListboxProc(hwnd, uMsg, wParam, lParam);
    }

}

    /*
    ###########################################################################
    # (@)INTERNAL:RichEditMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for subclassed Win32::GUI::RichEdit objects
    */
LRESULT CALLBACK RichEditMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(RichEditMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
#ifdef PERL_OBJECT
    CPerl *pPerl;
	pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
	if(pPerl == NULL) return DefRichEditProc(hwnd, uMsg, wParam, lParam);
#endif
    int PerlResult;
    char Name[MAX_EVENT_NAME];
    PerlResult = 1;
    strcpy((char *) Name, "main::");
    if(GetObjectName(NOTXSCALL hwnd, Name)) {
        switch(uMsg) {
        case WM_MOUSEMOVE:
            PerlResult = DoEvent_MouseMove(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_LBUTTONDOWN:
            strcat((char *) Name, "_LButtonDown");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_LBUTTONUP:
            strcat((char *) Name, "_LButtonUp");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_RBUTTONDOWN:
            strcat((char *) Name, "_RButtonDown");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_RBUTTONUP:
            strcat((char *) Name, "_RButtonUp");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_CHAR:
            strcat(Name, "_KeyPress");
            PerlResult = DoEvent_Long(NOTXSCALL Name, wParam);
            break;
        case WM_MOUSEHOVER:
        	strcat((char *) Name, "_MouseOver");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;
        case WM_MOUSELEAVE:
        	strcat((char *) Name, "_MouseOut");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;

        // to implement:
        // MouseUp
        // MouseDown
        }
    }
    if(PerlResult == 0) {
        return 0;
    } else {
        return DefRichEditProc(hwnd, uMsg, wParam, lParam);
    }

}

    /*
    ###########################################################################
    # (@)INTERNAL:TabStripMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for subclassed Win32::GUI::TabStrip objects
    */
LRESULT CALLBACK TabStripMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    // a TabStrip acts like a container, so we simply
    // redirect the messages to our parent and call the default Proc.
    HWND hwndParent;

    if(uMsg == WM_COMMAND || uMsg == WM_NOTIFY) {
        hwndParent = (HWND) GetWindowLong(hwnd, GWL_HWNDPARENT);
        SendMessage(hwndParent, uMsg, wParam, lParam);
        return 0;
    } else {
        return DefTabStripProc(hwnd, uMsg, wParam, lParam);
    }
}

    /*
    ###########################################################################
    # (@)INTERNAL:GraphicMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for Win32::GUI::Graphic objects
    */
LRESULT CALLBACK GraphicMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
#ifdef PERL_OBJECT
    CPerl *pPerl;
	pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
	if(pPerl == NULL) return DefWindowProc(hwnd, uMsg, wParam, lParam);
#endif
    int PerlResult;
    int count;
    char Name[MAX_EVENT_NAME];
	HV* self;
	SV* storing;
	SV* newdc;
	SV* newgraphic;
	HV* graphicclass;

    PerlResult = 1;

#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(GraphicMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
    if(uMsg == WM_PAINT) {
	    strcpy((char *) Name, "main::");
    	if(GetObjectName(NOTXSCALL hwnd, Name)) {
#ifdef PERLWIN32GUI_STRONGDEBUG
		    printf("!XS(GraphicMsgLoop) name=%s\n", Name);
#endif
			strcat((char *) Name, "_Paint");
    		if(perl_get_cv(Name, FALSE) != NULL) {
                /*
                 * (@)EVENT:Paint()
                 * Sent when the Graphic object needs to be repainted.
                 * Note that you need to use GetDC() to get the DC
                 * of the Graphic object where you do your paint
                 * work, and then Validate() the DC to inform Windows
                 * that you painted the DC area (otherwise it will
                 * continue to call the Paint event continuously).
                 * Example:
                 *   sub Graphic_Paint {
                 *       my $DC = $Window->Graphic->GetDC();
                 *       $DC->MoveTo(0, 0);
                 *       $DC->LineTo(100, 100);
                 *       $DC->Validate();
                 *   }
                 * (@)APPLIES_TO:Graphic
                 */

				/*
 				self = HV_SELF_FROM_WINDOW(hwnd);
				storing = newSViv((long) GetDC(hwnd));
 				hv_store_mg(NOTXSCALL self, "-handle", 7, storing, 0);
				storing = newSViv((long) hwnd);
 				hv_store_mg(NOTXSCALL self, "-window", 7, storing, 0);
                */

        		dSP;
        		dTARG;
        		ENTER;
        		SAVETMPS;
        		PUSHMARK(SP);
#ifdef PERLWIN32GUI_STRONGDEBUG
			    printf("!XS(GraphicMsgLoop) pushing parameters...\n");
#endif
				XPUSHs(sv_2mortal(newSVpv("Win32::GUI::DC", 0)));
#ifdef PERLWIN32GUI_STRONGDEBUG
			    printf("!XS(GraphicMsgLoop) done parameter 1...\n");
#endif
				XPUSHs(SV_SELF_FROM_WINDOW(hwnd));
#ifdef PERLWIN32GUI_STRONGDEBUG
			    printf("!XS(GraphicMsgLoop) done parameter 2...\n");
#endif
        		PUTBACK ;
#ifdef PERLWIN32GUI_STRONGDEBUG
			    printf("!XS(GraphicMsgLoop) doing perl_call...\n");
#endif
				count = perl_call_pv("Win32::GUI::DC::new", 0);
        		SPAGAIN ;
				newdc = newSVsv(POPs);
#ifdef PERLWIN32GUI_STRONGDEBUG
			    printf("!XS(GraphicMsgLoop) perl_call got(%d): %s\n", count, SvPV_nolen(newdc));
#endif
				PUTBACK;
				FREETMPS;
				LEAVE;

				ENTER;
				SAVETMPS;
	       		PUSHMARK(SP);
				XPUSHs(sv_2mortal(newdc));
				// XPUSHs(sv_2mortal(newdc));
#ifdef PERLWIN32GUI_STRONGDEBUG
			    printf("!XS(GraphicMsgLoop) doing perl_call_pv...\n");
#endif
				PUTBACK;
        		count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        		SPAGAIN;
         		if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            		if(count > 0) PerlResult = POPi;
        		}
        		PUTBACK;
        		FREETMPS;
        		LEAVE;
				/*
				storing = newSViv((long) hwnd);
 				hv_store_mg(NOTXSCALL self, "-handle", 7, storing, 0);
 				*/
    		}
    	}
    	return PerlResult;
    } else {
        return DefWindowProc(hwnd, uMsg, wParam, lParam);
    }
}


    /*
    ###########################################################################
    # (@)INTERNAL:InteractiveGraphicMsgLoop(hwnd, uMsg, wParam, lParam)
    # message loop for Win32::GUI::Graphic objects (with -interactive => 1
    # option).
    */
LRESULT CALLBACK InteractiveGraphicMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
#ifdef PERL_OBJECT
    CPerl *pPerl;
	pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
	if(pPerl == NULL) return DefWindowProc(hwnd, uMsg, wParam, lParam);
#endif
    int PerlResult;
    int count;
    char Name[MAX_EVENT_NAME];
	HV* self;
	SV* storing;
	SV* newdc;
	SV* newgraphic;
	HV* graphicclass;

    PerlResult = 1;

#ifdef PERLWIN32GUI_STRONGDEBUG
    printf("!XS(InteractiveGraphicMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif

	strcpy((char *) Name, "main::");

	if(GetObjectName(NOTXSCALL hwnd, Name)) {
		switch(uMsg) {
		case WM_PAINT:
			strcat((char *) Name, "_Paint");
    		if(perl_get_cv(Name, FALSE) != NULL) {
                /*
                 * (@)EVENT:Paint()
                 * Sent when the Graphic object needs to be repainted.
                 * Note that you need to use GetDC() to get the DC
                 * of the Graphic object where you do your paint
                 * work, and then Validate() the DC to inform Windows
                 * that you painted the DC area (otherwise it will
                 * continue to call the Paint event continuously).
                 * Example:
                 *   sub Graphic_Paint {
                 *       my $DC = $Window->Graphic->GetDC();
                 *       $DC->MoveTo(0, 0);
                 *       $DC->LineTo(100, 100);
                 *       $DC->Validate();
                 *   }
                 * (@)APPLIES_TO:Graphic
                 */
        		dSP;
        		dTARG;
        		ENTER;
        		SAVETMPS;
        		PUSHMARK(SP);
				XPUSHs(sv_2mortal(newSVpv("Win32::GUI::DC", 0)));
				XPUSHs(SV_SELF_FROM_WINDOW(hwnd));
        		PUTBACK ;
				count = perl_call_pv("Win32::GUI::DC::new", 0);
        		SPAGAIN ;
				newdc = newSVsv(POPs);
				PUTBACK;
				FREETMPS;
				LEAVE;

				ENTER;
				SAVETMPS;
	       		PUSHMARK(SP);
				XPUSHs(sv_2mortal(newdc));
				PUTBACK;
        		count = perl_call_pv(Name, G_EVAL|G_ARRAY);
        		SPAGAIN;
         		if(!ProcessEventError(NOTXSCALL Name, &PerlResult)) {
            		if(count > 0) PerlResult = POPi;
        		}
        		PUTBACK;
        		FREETMPS;
        		LEAVE;
			}
			break;
		case WM_MOUSEMOVE:
            PerlResult = DoEvent_MouseMove(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_LBUTTONDOWN:
            strcat((char *) Name, "_LButtonDown");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_LBUTTONUP:
            strcat((char *) Name, "_LButtonUp");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_RBUTTONDOWN:
            strcat((char *) Name, "_RButtonDown");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_RBUTTONUP:
            strcat((char *) Name, "_RButtonUp");
            PerlResult = DoEvent_MouseButton(NOTXSCALL Name, wParam, lParam);
            break;
        case WM_MOUSEHOVER:
        	strcat((char *) Name, "_MouseOver");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;
        case WM_MOUSELEAVE:
        	strcat((char *) Name, "_MouseOut");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
            if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
            break;
		}
	}
    if(PerlResult == -1) {
        PostMessage(hwnd, WM_EXITLOOP, -1, 0);
        return 0;
    } else {
        if(PerlResult == 0) {
            return 0;
        } else {
            return DefWindowProc(hwnd, uMsg, wParam, lParam);
        }
    }
}


    /*
    ###########################################################################
    # (@)INTERNAL:WindowMsgLoop(hwnd, uMsg, wParam, lParam)
    # this is the main message loop (WndProc)
    */
LRESULT CALLBACK WindowMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	dSP;
    int PerlResult;
    char Name[MAX_EVENT_NAME];
    int obj_class;
    LPNMHDR notify;
    LPNM_TREEVIEW tv_notify;
    TV_KEYDOWN FAR * tv_keydown;
	LPPERLWIN32GUI_USERDATA perlud;

#ifdef PERLWIN32GUI_STRONGDEBUG
	printf("!XS(WindowMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif
    if(uMsg == WM_CREATE || uMsg == WM_NCCREATE) {
        perlud = (LPPERLWIN32GUI_USERDATA) ((CREATESTRUCT *) lParam)->lpCreateParams;
        if(perlud!= NULL) {
            SetWindowLong(hwnd, GWL_USERDATA, (long) perlud);
        }
#ifdef PERLWIN32GUI_STRONGDEBUG
	printf("!XS(WindowMsgLoop) returning DefWindowProc (WM_CREATE)\n");
#endif
        return DefWindowProc(hwnd, uMsg, wParam, lParam);
    }
    if(uMsg == WM_DESTROY) {
		safefree( (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA));
#ifdef PERLWIN32GUI_STRONGDEBUG
		printf("!XS(NEM_ChildMsgLoop) returning DefWindowProc (WM_DESTROY)\n");
#endif
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}

#ifdef PERL_OBJECT
    CPerl *pPerl;
	pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
	if(pPerl == NULL) return DefWindowProc(hwnd, uMsg, wParam, lParam);
#endif

	ENTER;
	SAVETMPS;

    PerlResult = 1;
    strcpy(Name, "main::");

    switch(uMsg) {
    case WM_ACTIVATE:
        if(GetObjectName(NOTXSCALL hwnd, Name)) {
            if(LOWORD(wParam) == WA_INACTIVE) {
                /*
                 * (@)EVENT:Deactivate()
                 * Sent when the window is deactivated.
                 * (@)APPLIES_TO:Window, DialogBox
                 */
                strcat(Name, "_Deactivate");
            } else {
                /*
                 * (@)EVENT:Activate()
                 * Sent when the window is activated.
                 * (@)APPLIES_TO:Window, DialogBox
                 */
                strcat(Name, "_Activate");
            }
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
        }
        break;

    case WM_SYSCOMMAND:
        if(GetObjectName(NOTXSCALL hwnd, Name)) {
            switch(wParam & 0xFFF0) {
            case SC_CLOSE:
                /*
                 * (@)EVENT:Terminate()
                 * Sent when the window is closed.
                 * The event should return -1 to terminate the interaction
                 * and return control to the perl script; see Dialog().
                 * (@)APPLIES_TO:Window, DialogBox
                 */
                strcat(Name, "_Terminate");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case SC_MINIMIZE:
                /*
                 * (@)EVENT:Minimize()
                 * Sent when the window is minimized.
                 * (@)APPLIES_TO:Window, DialogBox
                 */
                strcat(Name, "_Minimize");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case SC_MAXIMIZE:
                /*
                 * (@)EVENT:Maximize()
                 * Sent when the window is maximized.
                 * (@)APPLIES_TO:Window, DialogBox
                 */
                strcat(Name, "_Maximize");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            }
        }
        break;

    case WM_SIZE:
        if(GetObjectName(NOTXSCALL hwnd, Name)) {
            /*
             * (@)EVENT:Resize()
             * Sent when the window is resized.
             * (@)APPLIES_TO:Window, DialogBox
             */
            strcat(Name, "_Resize");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
        }
        break;

    case WM_COMMAND:
        if(HIWORD(wParam) == 0 && lParam == NULL) {
            // menu command processing
#ifdef PERLWIN32GUI_STRONGDEBUG
				printf("!XS(WindowMsgLoop) got WM_COMMAND for a menu...\n");
#endif
            if(GetMenuName(NOTXSCALL hwnd, LOWORD(wParam), Name)) {
                /*
                 * (@)EVENT:Click()
				 * Sent when the users choose a menu point.
                 * (@)APPLIES_TO:Menu
                 */
                strcat(Name, "_Click");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
            }
        } else if(HIWORD(wParam) == 1 && lParam == NULL) {
            // accelerator processing
#ifdef PERLWIN32GUI_STRONGDEBUG
				printf("!XS(WindowMsgLoop) got WM_COMMAND for an accelerator...\n");
#endif
            if(GetAcceleratorName(NOTXSCALL LOWORD(wParam), Name)) {
                /*
                 * (@)EVENT:Click()
				 * Sent when the users triggers an Accelerator object.
                 * (@)APPLIES_TO:AcceleratorTable
                 */
                strcat(Name, "_Click");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
            }
        } else {
#ifdef PERLWIN32GUI_STRONGDEBUG
				printf("!XS(WindowMsgLoop) got WM_COMMAND, doing GetObjectNameAndClass...\n");
#endif
            if(GetObjectNameAndClass(NOTXSCALL (HWND) lParam, Name, &obj_class)) {
#ifdef PERLWIN32GUI_STRONGDEBUG
					printf("!XS(WindowMsgLoop) GetObjectNameAndClass succeeded (Name=%s, class=%d)...\n", Name, obj_class);
#endif
                switch(obj_class) {

                case WIN32__GUI__BUTTON:
                case WIN32__GUI__CHECKBOX:
                case WIN32__GUI__RADIOBUTTON:
                    switch(HIWORD(wParam)) {
                    case BN_SETFOCUS:
                        /*
                         * (@)EVENT:GotFocus()
						 * Sent when the control is activated.
                         * (@)APPLIES_TO:Button, Checkbox, RadioButton
                         */
                        strcat((char *) Name, "_GotFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case BN_KILLFOCUS:
                        /*
                         * (@)EVENT:LostFocus()
						 * Sent when the control is deactivated.
                         * (@)APPLIES_TO:Button, Checkbox, RadioButton
                         */
                        strcat((char *) Name, "_LostFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case BN_CLICKED:
                        /*
                         * (@)EVENT:Click()
						 * Sent when the control is selected (eg.
						 * the button pushed, the checkbox checked, etc.).
                         * (@)APPLIES_TO:Button, Checkbox, RadioButton
                         */
                        strcat((char *) Name, "_Click");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case BN_DBLCLK:
                        /*
                         * (@)EVENT:DblClick()
						 * Sent when the user double clicks on the control.
                         * (@)APPLIES_TO:Button, Checkbox, RadioButton
                         */
                        strcat((char *) Name, "_DblClick");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
			  		case BN_PUSHED:
                        /*
                         * (@)EVENT:MouseDown()
                         * Sent when the user down clicks on the control.
                         * (@)APPLIES_TO:Button, Checkbox, RadioButton
                         */
						strcat((char *) Name, "_MouseDown");
						PerlResult = DoEvent_Generic(NOTXSCALL Name);
						break;
					case BN_UNPUSHED:
                        /*
                         * (@)EVENT:MouseUp()
                         * Sent when the user releases a down click on the control.
                         * (@)APPLIES_TO:Button, Checkbox, RadioButton
                         */
						strcat((char *) Name, "_MouseUp");
						PerlResult = DoEvent_Generic(NOTXSCALL Name);
						break;
                    default:
                        strcat((char *) Name, "_Anonymous");
#ifdef PERLWIN32GUI_STRONGDEBUG
	                        printf("!XS(WindowMsgLoop): BUTTON WM_COMMAND NotifyCode=%d\n", HIWORD(wParam));
#endif
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
		    }
                    break;

                case WIN32__GUI__LISTBOX:
                    switch(HIWORD(wParam)) {
                    case LBN_SETFOCUS:
                        /*
                         * (@)EVENT:GotFocus()
						 * Sent when the control is activated.
                         * (@)APPLIES_TO:Listbox
                         */
                        strcat((char *) Name, "_GotFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case LBN_KILLFOCUS:
                        /*
                         * (@)EVENT:LostFocus()
						 * Sent when the control is deactivated.
						 * (@)APPLIES_TO:Listbox
                         */
                        strcat((char *) Name, "_LostFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case LBN_SELCHANGE:
                        /*
                         * (@)EVENT:Click()
                         * (@)APPLIES_TO:Listbox
                         */
                        strcat((char *) Name, "_Click");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case LBN_DBLCLK:
                        /*
                         * (@)EVENT:DblClick()
						 * Sent when the user double clicks on the control.
						 * (@)APPLIES_TO:Listbox
                         */
                        strcat((char *) Name, "_DblClick");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    default:
                        strcat((char *) Name, "_Anonymous");
#ifdef PERLWIN32GUI_STRONGDEBUG
	                        printf("!XS(WindowMsgLoop): LISTBOX WM_COMMAND NotifyCode=%d\n", HIWORD(wParam));
#endif
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    }
                    break;

                case WIN32__GUI__EDIT:
                case WIN32__GUI__RICHEDIT:
                    switch(HIWORD(wParam)) {
                    case EN_SETFOCUS:
                        /*
                         * (@)EVENT:GotFocus()
                         * Sent when the control is activated.
                         * (@)APPLIES_TO:Textfield, RichEdit
                         */
                        strcat((char *) Name, "_GotFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case EN_KILLFOCUS:
                        /*
                         * (@)EVENT:LostFocus()
                         * Sent when the control is deactivated.
                         * (@)APPLIES_TO:Textfield, RichEdit
                         */
                        strcat((char *) Name, "_LostFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case EN_CHANGE:
                        /*
                         * (@)EVENT:Change()
                         * Sent when the text in the field is changed by the user.
                         * (@)APPLIES_TO:Textfield, RichEdit
                         */
                        strcat((char *) Name, "_Change");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    default:
                        strcat((char *) Name, "_Anonymous");
#ifdef PERLWIN32GUI_STRONGDEBUG
	                        printf("!XS(WindowMsgLoop): EDIT WM_COMMAND NotifyCode=%d\n", HIWORD(wParam));
#endif
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    }
                    break;

                case WIN32__GUI__STATIC:
                    switch(HIWORD(wParam)) {
                    case STN_CLICKED:
                        /*
                         * (@)EVENT:Click()
                         * (@)APPLIES_TO:Label
                         */
                        strcat((char *) Name, "_Click");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case STN_DBLCLK:
                        /*
                         * (@)EVENT:DblClick()
                         * (@)APPLIES_TO:Label
                         */
                        strcat((char *) Name, "_DblClick");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    default:
                        strcat((char *) Name, "_Anonymous");
#ifdef PERLWIN32GUI_STRONGDEBUG
	                        printf("!XS(WindowMsgLoop): STATIC WM_COMMAND NotifyCode=%d\n", HIWORD(wParam));
#endif
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    }
                    break;

                case WIN32__GUI__COMBOBOX:
                case WIN32__GUI__COMBOBOXEX:
                    switch(HIWORD(wParam)) {
                    case CBN_SETFOCUS:
                        /*
                         * (@)EVENT:GotFocus()
						 * Sent when the control is activated.
                         * (@)APPLIES_TO:Combobox, ComboboxEx
                         */
                        strcat((char *) Name, "_GotFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    case CBN_KILLFOCUS:
                        /*
                         * (@)EVENT:LostFocus()
						 * Sent when the control is deactivated.
						 * (@)APPLIES_TO:Combobox, ComboboxEx
                         */
                        strcat((char *) Name, "_LostFocus");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
					case CBN_SELCHANGE:
                        /*
                         * (@)EVENT:Change()
                         * Sent when the user selects an item from the Combobox
                         * (@)APPLIES_TO:Combobox, ComboboxEx
                         */
                        strcat((char *) Name, "_Change");
                        PerlResult = DoEvent_Generic(NOTXSCALL Name);
                        break;
                    default:
                        strcat((char *) Name, "_Anonymous");
#ifdef PERLWIN32GUI_STRONGDEBUG
	                        printf("!XS(WindowMsgLoop): COMBOBOX WM_COMMAND NotifyCode=%d\n", HIWORD(wParam));
#endif
#ifdef PERLWIN32GUI_STRONGDEBUG
	                        printf("!XS(WindowMsgLoop): COMBOBOX WM_COMMAND hWnd=0x%x\n", lParam);
#endif
                        PerlResult = DoEvent_Long(NOTXSCALL Name, HIWORD(wParam));
                        break;
                    }
                    break;

                case WIN32__GUI__TOOLBAR:
                    /*
                     * (@)EVENT:ButtonClick(INDEX)
                     * Sent when the user presses a button of the Toolbar
                     * the INDEX argument identifies the zero-based index of
                     * the pressed button
                     * (@)APPLIES_TO:Toolbar
                     */
                    strcat((char *) Name, "_ButtonClick");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, LOWORD(wParam));
                    break;

                }
            }
        }
        break;


    case WM_NOTIFY:
        notify = (LPNMHDR) lParam;
        if(GetObjectNameAndClass(NOTXSCALL notify->hwndFrom, Name, &obj_class)) {
            switch(obj_class) {

            case WIN32__GUI__TOOLBAR:
                {
                    LPTBNOTIFY tbn;
                    tbn = (LPTBNOTIFY) lParam;
#ifdef PERLWIN32GUI_STRONGDEBUG
	                printf("!XS(WindowMsgLoop): TOOLBAR WM_NOTIFY Code=%ud\n", tbn->hdr.code);
#endif
                }
                break;

            case WIN32__GUI__LISTVIEW:
                {
                    LPNM_LISTVIEW lv_notify;
                    lv_notify = (LPNM_LISTVIEW) lParam;
                    switch(notify->code) {
                    case LVN_ITEMCHANGED:
                        if(lv_notify->uChanged & LVIF_STATE
                        && lv_notify->uNewState & LVIS_SELECTED) {
                            /*
                             * (@)EVENT:ItemClick(ITEM)
                             * Sent when the user selects an item in the ListView;
                             * ITEM specifies the zero-based index of the selected item.
                             * (@)APPLIES_TO:ListView
                             */
                            strcat((char *) Name, "_ItemClick");
                            PerlResult = DoEvent_Long(NOTXSCALL Name, lv_notify->iItem);
                        }
						if(lv_notify->uChanged & LVIF_STATE
						&& (lv_notify->uOldState & LVIS_STATEIMAGEMASK) != (lv_notify->uNewState & LVIS_STATEIMAGEMASK)) {
                            /*
                             * (@)EVENT:ItemCheck(ITEM)
                             * Sent when the user changes the checkbox of an item in the ListView;
                             * ITEM specifies the zero-based index of the selected item.
                             * (@)APPLIES_TO:ListView
                             */
                            strcat((char *) Name, "_ItemCheck");
                            PerlResult = DoEvent_Long(NOTXSCALL Name, lv_notify->iItem);
						}

                        break;
                    case LVN_COLUMNCLICK:
                        /*
                         * (@)EVENT:ColumnClick(ITEM)
                         * Sent when the user clicks on a column header in the
                         * ListView; ITEM specifies the one-based index of the
                         * selected column.
                         * (@)APPLIES_TO:ListView
                         */
                        strcat((char *) Name, "_ColumnClick");
                        PerlResult = DoEvent_Long(NOTXSCALL Name, lv_notify->iSubItem);
                        break;
                    case LVN_KEYDOWN:
                        {
                            LV_KEYDOWN FAR * lv_keydown;
                            lv_keydown = (LV_KEYDOWN FAR *) lParam;
                            /*
                             * (@)EVENT:KeyDown(KEY)
							 * Sent when the user presses a key while the ListView
							 * control has focus; KEY is the ASCII code of the
							 * key being pressed.
                             * (@)APPLIES_TO:ListView
                             */
                            strcat((char *) Name, "_KeyDown");
                            PerlResult = DoEvent_Long(NOTXSCALL Name, lv_keydown->wVKey);
                        }
                        break;
                    }
                }
                break;

            case WIN32__GUI__TREEVIEW:
                tv_notify = (LPNM_TREEVIEW) lParam;
                switch(notify->code) {
                case TVN_ITEMEXPANDED:
                    if(tv_notify->action == TVE_COLLAPSE)
                        /*
                         * (@)EVENT:Collapse(NODE)
                         * Sent when the user closes the specified NODE of the TreeView.
                         * (@)APPLIES_TO:TreeView
                         */
                        strcat(Name, "_Collapse");
                    else
                        /*
                         * (@)EVENT:Expand(NODE)
                         * Sent when the user opens the specified NODE of the TreeView.
                         * (@)APPLIES_TO:TreeView
                         */
                        strcat(Name, "_Expand");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, (long) tv_notify->itemNew.hItem);
                    break;

                case TVN_ITEMEXPANDING:
                    if(tv_notify->action == TVE_COLLAPSE)
                        /*
                         * (@)EVENT:Collapsing(NODE)
                         * Sent when the user is about to close the
                         * specified NODE of the TreeView.
                         * The event should return 0 to prevent the
                         * action, 1 to allow it.
                         * (@)APPLIES_TO:TreeView
                         */
                        strcat(Name, "_Collapsing");
                    else
                        /*
                         * (@)EVENT:Expanding(NODE)
                         * Sent when the user is about to open the
                         * specified NODE of the TreeView
                         * The event should return 0 to prevent the
                         * action, 1 to allow it.
                         * (@)APPLIES_TO:TreeView
                         */
                        strcat(Name, "_Expanding");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, (long) tv_notify->itemNew.hItem);
                    FREETMPS;
                    LEAVE;
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                    break;

                case TVN_SELCHANGED:
                    /*
                     * (@)EVENT:NodeClick(NODE)
                     * Sent when the user clicks on the specified NODE of the TreeView.
                     * (@)APPLIES_TO:TreeView
                     */
                    strcat(Name, "_NodeClick");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, (long) tv_notify->itemNew.hItem);
                    break;

                case TVN_KEYDOWN:
                    /*
                     * (@)EVENT:KeyDown(KEY)
					 * Sent when the user presses a key while the TreeView
					 * control has focus; KEY is the ASCII code of the
					 * key being pressed.
					 * (@)APPLIES_TO:TreeView
                     */
                    tv_keydown = (TV_KEYDOWN FAR *) lParam;
                    strcat((char *) Name, "_KeyDown");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, tv_keydown->wVKey);
                    break;
                }
                break;

            case WIN32__GUI__TAB:
                switch(notify->code) {
                case TCN_SELCHANGING:
                    /*
                     * (@)EVENT:Changing()
                     * Sent before the current selection changes.
					 * Use SelectedItem() to determine the
					 * current selection.
                     * The event should return 0 to prevent
                     * the selection changing, 1 to allow it.
                     * (@)APPLIES_TO:TabStrip
                     */
                    strcat((char *) Name, "_Changing");
                    PerlResult = DoEvent_Generic(NOTXSCALL Name);
                    FREETMPS;
                    LEAVE;
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                    break;
                case TCN_SELCHANGE:
                    /*
                     * (@)EVENT:Change()
                     * Sent when the current
                     * selection has changed. Use SelectedItem()
                     * to determine the current selection.
                     * (@)APPLIES_TO:TabStrip
				 	 */
                    strcat((char *) Name, "_Change");
                    PerlResult = DoEvent_Generic(NOTXSCALL Name);
                    break;
                }
                break;

            case WIN32__GUI__TOOLTIP:
                if(((LPNMHDR)lParam)->code == TTN_NEEDTEXT) {
                    /*
                     * (@)EVENT:NeedText(ID)
                     * (@)APPLIES_TO:Tooltip
                     */
                    LPTOOLTIPTEXT lptt;
                    lptt = (LPTOOLTIPTEXT) lParam;
                    lptt->lpszText = (LPTSTR) DoEvent_NeedText(NOTXSCALL Name, lptt->hdr.idFrom);
#ifdef PERLWIN32GUI_STRONGDEBUG
	                    printf("!XS(WindowMsgLoop): TTN_NEEDTEXT got '%s'\n", lptt->lpszText);
#endif
                    PerlResult = 1;

                }
                break;

            case WIN32__GUI__REBAR:
                if(((LPNMHDR)lParam)->code == RBN_HEIGHTCHANGE) {
                    /*
                     * (@)EVENT:HeightChange()
                     * Sent when the height of the Rebar control has changed.
                     * (@)APPLIES_TO:Rebar
                     */
                    strcat((char *) Name, "_HeightChange");
                    PerlResult = DoEvent_Generic(NOTXSCALL Name);
                }
                break;

            case WIN32__GUI__HEADER:
                if(((LPNMHDR)lParam)->code == HDN_BEGINTRACK) {
                    /*
                     * (@)EVENT:BeginTrack(INDEX, WIDTH)
                     * Sent when a divider of the Header control
                     * is being moved; the event must return 0 to
                     * prevent moving the divider, 1 to allow it.
                     * Passes the zero-based INDEX
                     * of the item being resized and its current
                     * WIDTH.
                     * (@)APPLIES_TO:Header
                     */
                    LPNMHEADER nmh;
                    nmh = (LPNMHEADER) lParam;
                    strcat((char *) Name, "_BeginTrack");
                    PerlResult = DoEvent_TwoLongs(
                        NOTXSCALL Name,
                        nmh->iItem,
                        nmh->pitem->cxy
                    );
                    FREETMPS;
                    LEAVE;
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                } else if(((LPNMHDR)lParam)->code == HDN_ENDTRACK) {
                    /*
                     * (@)EVENT:EndTrack(INDEX, WIDTH)
                     * Sent when a divider of the Header control
                     * has been moved. Passes the zero-based INDEX
                     * of the item being resized and its current
                     * WIDTH.
                     * (@)APPLIES_TO:Header
                     */
                    LPNMHEADER nmh;
                    nmh = (LPNMHEADER) lParam;
                    strcat((char *) Name, "_EndTrack");
                    PerlResult = DoEvent_TwoLongs(
                        NOTXSCALL Name,
                        nmh->iItem,
                        nmh->pitem->cxy
                    );
                } else if(((LPNMHDR)lParam)->code == HDN_TRACK) {
                    /*
                     * (@)EVENT:Track(INDEX, WIDTH)
                     * Sent while a divider of the Header control
                     * is being moved; the event must return 1 to
                     * continue moving the divider, 0 to end its
                     * movement.
                     * Passes the zero-based INDEX
                     * of the item being resized and its current
                     * WIDTH.
                     * (@)APPLIES_TO:Header
                     */
                    LPNMHEADER nmh;
                    nmh = (LPNMHEADER) lParam;
                    strcat((char *) Name, "_Track");
                    PerlResult = DoEvent_TwoLongs(
                        NOTXSCALL Name,
                        nmh->iItem,
                        nmh->pitem->cxy
                    );
                    FREETMPS;
                    LEAVE;
                    if(PerlResult == 0) return TRUE;
                    else                return FALSE;
                } else if(((LPNMHDR)lParam)->code == HDN_DIVIDERDBLCLICK) {
                    /*
                     * (@)EVENT:DividerDblClick(INDEX)
                     * Sent when the user double-clicked on a
                     * divider of the Header control.
                     * (@)APPLIES_TO:Header
                     */
                    LPNMHEADER nmh;
                    nmh = (LPNMHEADER) lParam;
                    strcat((char *) Name, "_DividerDblClick");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, nmh->iItem);
                } else if(((LPNMHDR)lParam)->code == HDN_ITEMCLICK) {
                    /*
                     * (@)EVENT:ItemClick(INDEX)
                     * Sent when the user clicked on a Header
                     * item.
                     * (@)APPLIES_TO:Header
                     */
                    LPNMHEADER nmh;
                    nmh = (LPNMHEADER) lParam;
                    strcat((char *) Name, "_ItemClick");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, nmh->iItem);
                } else if(((LPNMHDR)lParam)->code == HDN_ITEMDBLCLICK) {
                    /*
                     * (@)EVENT:ItemDblClick(INDEX)
                     * Sent when the user double-clicked on a Header
                     * item.
                     * (@)APPLIES_TO:Header
                     */
                    LPNMHEADER nmh;
                    nmh = (LPNMHEADER) lParam;
                    strcat((char *) Name, "_ItemDblClick");
                    PerlResult = DoEvent_Long(NOTXSCALL Name, nmh->iItem);
                }
                break;

            }
            /*
             * ###############################################
             * standard notifications (true for all controls?)
             * ###############################################
             */
            switch(notify->code) {
            case NM_CLICK:
                /*
                 * (@)EVENT:Click()
                 * (@)APPLIES_TO:*
                 */
                strcat((char *) Name, "_Click");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case NM_RCLICK:
                /*
                 * (@)EVENT:RightClick()
                 * (@)APPLIES_TO:*
                 */
                strcat((char *) Name, "_RightClick");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case NM_DBLCLK:
                /*
                 * (@)EVENT:DblClick()
                 * (@)APPLIES_TO:*
                 */
                strcat((char *) Name, "_DblClick");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case NM_RDBLCLK:
                /*
                 * (@)EVENT:DblRightClick()
                 * (@)APPLIES_TO:*
                 */
                strcat((char *) Name, "_DblRightClick");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case NM_SETFOCUS:
                /*
                 * (@)EVENT:GotFocus()
                 * (@)APPLIES_TO:*
                 */
                strcat((char *) Name, "_GotFocus");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case NM_KILLFOCUS:
                /*
                 * (@)EVENT:LostFocus()
                 * (@)APPLIES_TO:*
                 */
                strcat((char *) Name, "_LostFocus");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            }
        }
        break;

    case WM_TIMER:
        /* (@)EVENT:Timer() */
        if(GetTimerName(NOTXSCALL hwnd, wParam, Name)) {
            strcat((char *) Name, "_Timer");
            PerlResult = DoEvent_Generic(NOTXSCALL Name);
        } else {
            PerlResult = 0;
        }
        break;

    case WM_HSCROLL:
    case WM_VSCROLL:
        if(GetObjectNameAndClass(NOTXSCALL (HWND) lParam, Name, &obj_class)) {
            switch(obj_class) {
            case WIN32__GUI__TRACKBAR:
                /*
                 * (@)EVENT:Scroll()
                 * Sent when the user moves the slider handle.
                 * (@)APPLIES_TO:Slider
                 */
                strcat((char *) Name, "_Scroll");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case WIN32__GUI__UPDOWN:
                /*
                 * (@)EVENT:Scroll()
                 * Sent when the user presses either the up or down button
                 * of the UpDown control.
                 * (@)APPLIES_TO:UpDown
                 */
                strcat((char *) Name, "_Scroll");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            }
        }
        break;

	case WM_MOUSEHOVER:
        if(GetObjectName(NOTXSCALL hwnd, Name)) {
			strcat((char *) Name, "_MouseOver");
			PerlResult = DoEvent_Generic(NOTXSCALL Name);
			if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
		}
		break;
	case WM_MOUSELEAVE:
		if(GetObjectName(NOTXSCALL hwnd, Name)) {
			strcat((char *) Name, "_MouseOut");
			PerlResult = DoEvent_Generic(NOTXSCALL Name);
			if(PerlResult) {
				TRACKMOUSEEVENT tme;
				tme.cbSize = sizeof(TRACKMOUSEEVENT);
				tme.hwndTrack = hwnd;
				tme.dwFlags = TME_QUERY;
				if(TrackMouseEvent( &tme )) {
					TrackMouseEvent( &tme );
				}
			}
		}
		break;


/*
    case WM_MOUSEMOVE:
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
        {
            HV* self;
            SV** rtts;
            AV* tts;
            I32 ttsi;
            SV** rtt;
            HWND tt;
            MSG ttmsg;
            self = HV_SELF_FROM_WINDOW(hwnd);
            if(self != NULL) {
                rtts = hv_fetch_mg(NOTXSCALL self, "-tooltips", 9, FALSE);
                if(rtts != NULL && SvOK(*rtts)) {
                	if(SvROK(*rtts) && SvTYPE(SvRV(*rtts)) == SVt_PVAV) {
	                    tts = (AV*) SvRV(*rtts);
#ifdef PERLWIN32GUI_STRONGDEBUG
		                printf("!XS(WindowMsgLoop): found -tooltips (%d)...\n", av_len(tts));
#endif
	                    for(ttsi=0;ttsi<av_len(tts);ttsi++) {
                        	rtt = av_fetch(tts, ttsi, 0);
                        	if(rtt != NULL) {
	                            tt = (HWND) SvIV(*rtt);
#ifdef PERLWIN32GUI_STRONGDEBUG
		                        printf("!XS(WindowMsgLoop): relaying to tooltip 0x%x...\n", tt);
#endif
	                            ttmsg.hwnd = hwnd;
                            	ttmsg.lParam = lParam;
                            	ttmsg.wParam = wParam;
                            	ttmsg.message = uMsg;
                            	SendMessage(tt, TTM_RELAYEVENT, 0, (LPARAM) &ttmsg);
                        	}
                        }
                    }
                }
            }
        }
        break;
*/
    case WM_NOTIFYICON:
        if(GetNotifyIconName(NOTXSCALL hwnd, (UINT) wParam, Name)) {
            switch(lParam) {
            case WM_LBUTTONDOWN:
                /*
                 * (@)EVENT:Click()
                 * Sent when the user clicks the left mouse button on
                 * a NotifyIcon.
                 * (@)APPLIES_TO:NotifyIcon
                 */
                strcat(Name, "_Click");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            case WM_RBUTTONDOWN:
                /*
                 * (@)EVENT:RightClick()
                 * Sent when the user clicks the right mouse button on
                 * a NotifyIcon.
                 * (@)APPLIES_TO:NotifyIcon
                 */
                strcat(Name, "_RightClick");
                PerlResult = DoEvent_Generic(NOTXSCALL Name);
                break;
            default:
                /*
                 * (@)EVENT:MouseEvent(MSG)
                 * Sent when the user performs a mouse event on
                 * a NotifyIcon; MSG is the message code.
                 * (@)APPLIES_TO:NotifyIcon
                 */
                strcat(Name, "_MouseEvent");
                PerlResult = DoEvent_Long(NOTXSCALL Name, (long)lParam);
                break;
            }
        }
        break;
	}
	FREETMPS;
	LEAVE;

    if(PerlResult == -1) {
        PostMessage(hwnd, WM_EXITLOOP, -1, 0);
        return 0;
    } else {
        if(PerlResult == 0) {
            return 0;
        } else {
			return CommonMsgLoop(NOTXSCALL hwnd, uMsg, wParam, lParam);
        }
    }

}

    /*
    ###########################################################################
    # (@)INTERNAL:MsgLoop(hwnd, uMsg, wParam, lParam)
    # obsolete (?) Win32::GUI::Window message loop
    */
LRESULT CALLBACK MsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
#ifdef PERL_OBJECT
    CPerl *pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
#endif
    HV* cb_hash;
    SV** cb_ref;
    int i;
    int *Items;
    int Count;
    int PerlResult = 1;
    char temp[80];
    int send_to = 0;
    // 0 = DefWindowProc
    // 1 = function: perl_call_sv(cb_ref...
    // 2 = function or DefWindowProc

    dSP;
    dTARG;

    ENTER ;
    SAVETMPS;

    PUSHMARK(SP) ;
    EXTEND(SP,4) ;
    PUSHs(sv_2mortal(newSViv((long)hwnd)));
    PUSHs(sv_2mortal(newSViv((long)uMsg)));
    PUSHs(sv_2mortal(newSViv((long)wParam)));
    PUSHs(sv_2mortal(newSViv((long)lParam)));
    PUTBACK ;

    if(uMsg == WM_COMMAND) {

        if(lParam == NULL) {
            // Menu Option
            if(HIWORD(wParam) == 0) {
                cb_hash = perl_get_hv("Win32::GUI::menucallbacks", FALSE);
                ltoa((long) LOWORD(wParam), temp, 10);
                if(hv_exists(cb_hash, temp, strlen(temp))) {
                    cb_ref = hv_fetch(cb_hash, temp, strlen(temp), FALSE);
                    send_to = 1; // cb_ref
                } else {
                    send_to = 2;
                }
            } else {
                send_to = 2; // ...hwnd or Def
            }
        } else {
            cb_hash = perl_get_hv("Win32::GUI::callbacks", FALSE);
            ltoa((long)lParam, temp, 10);
            if(hv_exists(cb_hash, temp, strlen(temp))) {
                cb_ref = hv_fetch(cb_hash, temp, strlen(temp), FALSE);
                send_to = 1; // cb_ref
            } else {
                send_to = 2; // ...hwnd or Def
            }
        }
    } else if(uMsg == WM_NOTIFY) {
        LPNMHDR nmhdr = (LPNMHDR) lParam;
        HWND hwnd = nmhdr->hwndFrom;
        UINT id = nmhdr->idFrom;
        UINT code = nmhdr->code;
        cb_hash = perl_get_hv("Win32::GUI::callbacks", FALSE);
        ltoa((long)hwnd, temp, 10);
        if(hv_exists(cb_hash, temp, strlen(temp))) {
            cb_ref = hv_fetch(cb_hash, temp, strlen(temp), FALSE);
            send_to = 1; // cb_ref
        } else {
            send_to = 2; // ...hwnd or Def
        }
    } else {
        send_to = 2;
    }

    switch(send_to) {
    case 2:
        cb_hash = perl_get_hv("Win32::GUI::callbacks", FALSE);
        ltoa((long)hwnd, temp, 10);
        if(hv_exists(cb_hash, temp, strlen(temp))) {

            cb_ref = hv_fetch(cb_hash, temp, strlen(temp), FALSE);

            perl_call_pv((char *)SvPV_nolen(*cb_ref), G_ARRAY);

            SPAGAIN ;
            PerlResult = POPi;
            PUTBACK ;
        }
        break;
    case 1:
        perl_call_pv((char *)SvPV_nolen(*cb_ref), G_ARRAY);
        SPAGAIN ;
        PerlResult = POPi;
        PUTBACK ;
        break;
    }
    FREETMPS ;
    LEAVE ;

    if(PerlResult == -1) {
#ifdef PERLWIN32GUI_DEBUG
        printf("!XS(MsgLoop): posting WM_EXITLOOP to 0x%x...\n", hwnd);
#endif
        PostMessage(hwnd, WM_EXITLOOP, -1, 0);
        return 0;
    } else {
        if(PerlResult == 0) {
            return 0;
        } else {
            return DefWindowProc(hwnd, uMsg, wParam, lParam);
        }
    }
}

    /*
    ###########################################################################
    # (@)INTERNAL:InactiveMsgLoop(hwnd, uMsg, wParam, lParam)
    # called by DoModal
    */
LRESULT CALLBACK InactiveMsgLoop(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	dSP;
    int PerlResult;
    char Name[MAX_EVENT_NAME];
    int obj_class;
    LPNMHDR notify;
    LPNM_TREEVIEW tv_notify;
    TV_KEYDOWN FAR * tv_keydown;
	LPPERLWIN32GUI_USERDATA perlud;

#ifdef PERLWIN32GUI_STRONGDEBUG
	printf("!XS(InactiveMsgLoop) got (0x%x, 0x%x, 0x%x, 0x%x)\n", hwnd, uMsg, wParam, lParam);
#endif

#ifdef PERL_OBJECT
    CPerl *pPerl;
	pPerl = PERL_OBJECT_FROM_WINDOW(hwnd);
	if(pPerl == NULL) return DefWindowProc(hwnd, uMsg, wParam, lParam);
#endif

	perlud = (LPPERLWIN32GUI_USERDATA) GetWindowLong(hwnd, GWL_USERDATA);
	if( ValidUserData(perlud) ) {
		if(uMsg == WM_ACTIVATE && LOWORD(wParam) == WA_ACTIVE) {
			SendMessage(perlud->Modal, WM_ACTIVATE, wParam, (LPARAM) perlud->Modal);
			return 0;
		}
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}