
	/*
    ###########################################################################
    # (@)PACKAGE:Win32::GUI::Listbox
    ###########################################################################
	*/

#include "GUI.h"

BOOL
Listbox_onParseOption(NOTXSPROC char *option, SV* value, LPPERLWIN32GUI_CREATESTRUCT perlcs) {
	BOOL retval;
	retval = FALSE;
	if(strcmp(option, "-multisel") == 0) {
		retval = TRUE;
		if(SvIV(value) == 0) {
			SwitchBit(perlcs->cs.style, LBS_MULTIPLESEL, 0);
			SwitchBit(perlcs->cs.style, LBS_EXTENDEDSEL, 0);
		} else if(SvIV(value) == 1) {
			SwitchBit(perlcs->cs.style, LBS_MULTIPLESEL, 1);
			SwitchBit(perlcs->cs.style, LBS_EXTENDEDSEL, 0);
		} else if(SvIV(value) == 2) {
			SwitchBit(perlcs->cs.style, LBS_MULTIPLESEL, 1);
			SwitchBit(perlcs->cs.style, LBS_EXTENDEDSEL, 1);
		} else {
			if(PL_dowarn) warn("Win32::GUI: Invalid value for -multisel!");
		}
	} else BitmaskOptionValue("-sort", perlcs->cs.style, LBS_SORT)
	}
	return retval;
}


MODULE = Win32::GUI::Listbox		PACKAGE = Win32::GUI::Listbox

PROTOTYPES: DISABLE


    ###########################################################################
    # (@)METHOD:AddString(STRING)
LRESULT
AddString(handle,string)
    HWND handle
    LPCTSTR string
CODE:
    RETVAL = SendMessage(handle, LB_ADDSTRING, 0, (LPARAM) string);
OUTPUT:
    RETVAL

	###########################################################################
    # (@)METHOD:Add(STRING, STRING .. STRING)
void
Add(handle,...)
    HWND handle
PREINIT:
	int i;
	LRESULT res;
CODE:
	for(i = 1; i < items; i++) {
		SendMessage(handle, LB_ADDSTRING, 0, (LPARAM) (LPCTSTR) SvPV_nolen(ST(i)));
	}

    ###########################################################################
    # (@)METHOD:InsertItem(STRING, [INDEX])
    # Inserts an item at the specified zero-based INDEX in the Listbox,
    # or adds it at the end if INDEX is not specified.
LRESULT
InsertItem(handle,string,index=-1)
    HWND handle
    LPCTSTR string
    WPARAM index
CODE:
    RETVAL = SendMessage(handle, LB_INSERTSTRING, index, (LPARAM) string);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:GetString(INDEX)
    # Returns the string at the specified zero-based INDEX in the Listbox.
void
GetString(handle,index)
    HWND handle
    WPARAM index
PREINIT:
    STRLEN cbString;
    char *szString;
PPCODE:
    cbString = SendMessage(handle, LB_GETTEXTLEN, index, 0);
	if(cbString != LB_ERR) {
		szString = (char *) safemalloc(cbString);
		if(SendMessage(handle, LB_GETTEXT,
					   index, (LPARAM) (LPCTSTR) szString) != LB_ERR) {
			EXTEND(SP, 1);
			XST_mPV(0, szString);
			safefree(szString);
			XSRETURN(1);
		} else {
			safefree(szString);
			XSRETURN_UNDEF;
		}
	} else {
		XSRETURN_UNDEF;
	}

    ###########################################################################
    # (@)METHOD:ItemHeight([HEIGHT])
    # Gets or sets the items height in a Listbox.
LRESULT
ItemHeight(handle,height=-1)
    HWND handle
    long height
CODE:
    if(items == 1) {
        RETVAL = SendMessage(handle, LB_GETITEMHEIGHT, 0, 0);
    } else {
        RETVAL = SendMessage(handle, LB_SETITEMHEIGHT, 0, MAKELPARAM(height, 0));
    }
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:FirstVisibleItem([INDEX])
LRESULT
FirstVisibleItem(handle,index=-1)
    HWND handle
    WPARAM index
CODE:
    if(items == 1)
        RETVAL = SendMessage(handle, LB_GETTOPINDEX, 0, 0);
    else
        RETVAL = SendMessage(handle, LB_SETTOPINDEX, index, 0);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:ItemFromPoint(X, Y)
void
ItemFromPoint(handle,x,y)
    HWND handle
    UINT x
    UINT y
PREINIT:
    LRESULT lresult;
PPCODE:
    lresult = SendMessage(handle, LB_ITEMFROMPOINT, 0, (LPARAM) MAKELPARAM(x, y));
    if(GIMME == G_ARRAY) {
        EXTEND(SP, 2);
        XST_mIV(0, (long) LOWORD(lresult));
        if(HIWORD(lresult) == 0)
            XST_mIV(1, 1);
        else
            XST_mIV(1, 0);
        XSRETURN(2);
    } else {
        XSRETURN_IV((long) LOWORD(lresult));
    }


    ###########################################################################
    # (@)METHOD:SelectString(STRING, [INDEX])
LRESULT
SelectString(handle,string,index=-1)
    HWND handle
    LPCTSTR string
    WPARAM index
CODE:
    RETVAL = SendMessage(handle, LB_SELECTSTRING, index, (LPARAM) string);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:FindString(STRING, [INDEX])
LRESULT
FindString(handle,string,index=-1)
    HWND handle
    LPCTSTR string
    WPARAM index
CODE:
    RETVAL = SendMessage(handle, LB_FINDSTRING, index, (LPARAM) string);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:FindStringExact(STRING, [INDEX])
LRESULT
FindStringExact(handle,string,index=-1)
    HWND handle
    LPCTSTR string
    WPARAM index
CODE:
    RETVAL = SendMessage(handle, LB_FINDSTRINGEXACT, index, (LPARAM) string);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:SelectCount()
LRESULT
SelectCount(handle)
    HWND handle
CODE:
    RETVAL = SendMessage(handle, LB_GETSELCOUNT, 0, 0);
OUTPUT:
    RETVAL

    ###########################################################################
    # (@)METHOD:SelectedItems()
    # Returns an array containing the zero-based indexes of the selected items
    # in a multiple selection Listbox.
void
SelectedItems(handle)
    HWND handle
PREINIT:
    LRESULT count;
    LRESULT lresult;
    LPINT selitems;
    int i;
PPCODE:
    count = SendMessage(handle, LB_GETSELCOUNT, 0, 0);
    if(count > 0) {
        selitems = (LPINT) safemalloc(sizeof(INT)*count);
        lresult = SendMessage(handle, LB_GETSELITEMS, (WPARAM) count, (LPARAM) selitems);
        if(lresult == -1) {
            XSRETURN_UNDEF;
        } else {
            EXTEND(SP, lresult);
            for(i=0; i<lresult; i++) {
                XST_mIV(i, (long) selitems[i]);
            }
            safefree(selitems);
            XSRETURN(lresult);
        }
    } else {
        XSRETURN_UNDEF;
    }
