
#include "GUI.h"

	/*
    ###########################################################################
    # (@)PACKAGE:Win32::GUI::TabStrip
    ###########################################################################
	*/

BOOL
TabStrip_onParseOption(NOTXSPROC char *option, SV* value, LPPERLWIN32GUI_CREATESTRUCT perlcs) {
	BOOL retval;
	retval = FALSE;
	if(strcmp(option, "-imagelist") == 0) {
		retval = TRUE;
		perlcs->hImageList = (HIMAGELIST) handle_From(NOTXSCALL value);
	} else if(strcmp(option, "-vertical") == 0) {
		retval = TRUE;
		SwitchBit(perlcs->cs.style, TCS_VERTICAL, SvIV(value));
		SwitchBit(perlcs->cs.style, TCS_MULTILINE, SvIV(value));
	} else BitmaskOptionValue("-multiline",  perlcs->cs.style, TCS_MULTILINE)
	} else BitmaskOptionValue("-bottom",     perlcs->cs.style, TCS_BOTTOM)
	} else BitmaskOptionValue("-alignright", perlcs->cs.style, TCS_RIGHT)
	} else BitmaskOptionValue("-hottrack",   perlcs->cs.style, TCS_HOTTRACK)
	} else BitmaskOptionValue("-buttons",    perlcs->cs.style, TCS_BUTTONS)
	} else BitmaskOptionValue("-justify",    perlcs->cs.style, TCS_RIGHTJUSTIFY)
	} else BitmaskOptionValue("-flat",       perlcs->cs.style, TCS_FLATBUTTONS)
	}
	return retval;
}

void
TabStrip_onPostCreate(NOTXSPROC HWND myhandle, LPPERLWIN32GUI_CREATESTRUCT perlcs) {
	if(perlcs->hImageList != NULL)
		TabCtrl_SetImageList(myhandle, perlcs->hImageList);
}


MODULE = Win32::GUI::TabStrip		PACKAGE = Win32::GUI::TabStrip

    ###########################################################################
    # (@)METHOD:InsertItem(%OPTIONS)
    # Adds an item to the TabStrip.
    # Allowed %OPTIONS are:
    #  -image => NUMBER
    #    the index of an image from the associated ImageList
    #  -index => NUMBER
    #    the position for the new item (if not specified, the item
    #    is added at the end of the control)
    #  -text  => STRING
    #    the text that will appear on the item
int
InsertItem(handle,...)
    HWND handle
PREINIT:
    TC_ITEM Item;
    int iIndex;
    int iText;
    unsigned int chText;
    int i, next_i;
CODE:
    ZeroMemory(&Item, sizeof(TC_ITEM));
    iIndex = TabCtrl_GetItemCount(handle)+1;
    next_i = -1;
    for(i = 1; i < items; i++) {
        if(next_i == -1) {
            if(strcmp(SvPV_nolen(ST(i)), "-image") == 0) {
                next_i = i + 1;
                Item.mask = Item.mask | TCIF_IMAGE;
                Item.iImage = SvIV(ST(next_i));
            }
            if(strcmp(SvPV_nolen(ST(i)), "-index") == 0) {
                next_i = i + 1;
                iIndex = (int) SvIV(ST(next_i));
            }
            if(strcmp(SvPV_nolen(ST(i)), "-text") == 0) {
                next_i = i + 1;
                Item.pszText = SvPV(ST(next_i), chText);
                Item.cchTextMax = (int) chText;
                Item.mask = Item.mask | TCIF_TEXT;
            }
        } else {
            next_i = -1;
        }
    }
    RETVAL = TabCtrl_InsertItem(handle, iIndex, &Item);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:ChangeItem(ITEM, %OPTIONS)
    # Change most of the options used when the item was created
    # (see InsertItem()).
    # Allowed %OPTIONS are:
    #     -image
    #     -text
BOOL
ChangeItem(handle,item,...)
    HWND handle
    int item
PREINIT:
    TC_ITEM Item;
    int iIndex;
    int iText;
    unsigned int chText;
    int i, next_i;
CODE:
    ZeroMemory(&Item, sizeof(TC_ITEM));
    next_i = -1;
    for(i = 2; i < items; i++) {
        if(next_i == -1) {
            if(strcmp(SvPV_nolen(ST(i)), "-image") == 0) {
                next_i = i + 1;
                Item.mask = Item.mask | TCIF_IMAGE;
                Item.iImage = SvIV(ST(next_i));
            }
            if(strcmp(SvPV_nolen(ST(i)), "-text") == 0) {
                next_i = i + 1;
                Item.pszText = SvPV(ST(next_i), chText);
                Item.cchTextMax = (int) chText;
                Item.mask = Item.mask | TCIF_TEXT;
            }
        } else {
            next_i = -1;
        }
    }
    RETVAL = TabCtrl_SetItem(handle, item, &Item);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:Count()
    # Returns the number of items in the TabStrip.
int
Count(handle)
    HWND handle
CODE:
    RETVAL = TabCtrl_GetItemCount(handle);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:Reset()
    # Deletes all items from the TabStrip.
BOOL
Reset(handle)
    HWND handle
CODE:
    RETVAL = TabCtrl_DeleteAllItems(handle);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:DeleteItem(ITEM)
    # Removes the specified ITEM from the TabStrip.
BOOL
DeleteItem(handle,item)
    HWND handle
    int item
CODE:
    RETVAL = TabCtrl_DeleteItem(handle, item);
OUTPUT:
    RETVAL


    ###########################################################################
    # (@)METHOD:GetString(ITEM)
    # Returns the string associated with the specified ITEM in the TabStrip.
void
GetString(handle,item)
    HWND handle
    int item
PREINIT:
    char *szString;
    TC_ITEM tcItem;
PPCODE:
    szString = (char *) safemalloc(1024);
    tcItem.pszText = szString;
    tcItem.cchTextMax = 1024;
    tcItem.mask = TCIF_TEXT;
    if(TabCtrl_GetItem(handle, item, &tcItem)) {
        EXTEND(SP, 1);
        XST_mPV(0, szString);
        safefree(szString);
        XSRETURN(1);
    } else {
        safefree(szString);
        XSRETURN_UNDEF;
    }


      ################################################
	  # (@)METHOD:AdjustRect(LEFT, TOP, RIGHT, BOTTOM, [FLAG])
void
AdjustRect(handle,left,top,right,bottom,flag=0)
    HWND handle
    int left
    int top
    int right
    int bottom
    BOOL flag
PREINIT:
    RECT myRect;
PPCODE:
    myRect.left   = left;
    myRect.top    = top;
    myRect.right  = right;
    myRect.bottom = bottom;
    TabCtrl_AdjustRect(handle, flag, &myRect);
    EXTEND(SP, 4);
    XST_mIV(0, myRect.left);
    XST_mIV(1, myRect.top);
    XST_mIV(2, myRect.right);
    XST_mIV(3, myRect.bottom);
    XSRETURN(4);

      ################################################
      # (@)METHOD:Padding(X,Y)
void
Padding(handle,x,y)
	HWND handle
	int x
	int y
CODE:
	TabCtrl_SetPadding(handle, x, y);


      ################################################
      # (@)METHOD:MinTabWidth(WIDTH)
int
MinTabWidth(handle,width)
	HWND handle
	int width
CODE:
	RETVAL = TabCtrl_SetMinTabWidth(handle, width);
OUTPUT:
	RETVAL


      ################################################
      # (@)METHOD:RowCount()
int
RowCount(handle)
	HWND handle
CODE:
	RETVAL = TabCtrl_GetRowCount(handle);
OUTPUT:
	RETVAL


      ################################################
      # (@)METHOD:HitTest()
void
HitTest(handle,x,y)
	HWND handle
	int x
	int y
PREINIT:
	TCHITTESTINFO ht;
	int index;
PPCODE:
	ZeroMemory(&ht, sizeof(TCHITTESTINFO));
	ht.pt.x = x;
	ht.pt.y = y;
	index = TabCtrl_HitTest(handle, &ht);
	if(GIMME == G_ARRAY) {
		EXTEND(SP, 2);
		XST_mIV(0, index);
		XST_mIV(1, ht.flags);
		XSRETURN(2);
	} else {
		XSRETURN_IV(index);
	}

      ################################################
      # (@)METHOD:GetItemRect()
void
GetItemRect(handle,index)
	HWND handle
	int index
PREINIT:
	RECT myRect;
PPCODE:
	ZeroMemory(&myRect, sizeof(RECT));
	if(TabCtrl_GetItemRect(handle, index, &myRect)) {
		EXTEND(SP, 4);
		XST_mIV(0, myRect.left);
		XST_mIV(1, myRect.top);
		XST_mIV(2, myRect.right);
		XST_mIV(3, myRect.bottom);
		XSRETURN(4);
	} else {
		XSRETURN_UNDEF;
	}
